# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

import datetime
from collections.abc import Iterable, Iterator
from typing import cast

import pytest

from odfdo.cell import Cell
from odfdo.column import Column
from odfdo.document import Document
from odfdo.element import Element
from odfdo.header import Header
from odfdo.row import Row
from odfdo.table import Table, _get_python_value


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield cast(Table, document.body.get_table(name="Example1"))


def test_get_python_value_bytes():
    src = "éù".encode()
    assert isinstance(src, bytes)
    expected = "éù"
    assert _get_python_value(src, "utf8") == expected


def test_get_python_value_int():
    assert _get_python_value("42", "utf8") == 42


def test_get_python_value_int_2():
    assert _get_python_value(42, "utf8") == 42


def test_get_python_value_float():
    assert _get_python_value("42.123", "utf8") == 42.123


def test_get_python_value_float_2():
    assert _get_python_value(42.123, "utf8") == 42.123


def test_get_python_value_bool_true():
    assert _get_python_value(True, "utf8") is True


def test_get_python_value_bool_true_2():
    assert _get_python_value("true", "utf8") is True


def test_get_python_value_bool_true_3():
    assert _get_python_value("True", "utf8") is True


def test_get_python_value_bool_false():
    assert _get_python_value(False, "utf8") is False


def test_get_python_value_bool_false_2():
    assert _get_python_value("false", "utf8") is False


def test_get_python_value_bool_false_3():
    assert _get_python_value("False", "utf8") is False


def test_get_python_value_date():
    value = "2025-01-02"
    expected = datetime.datetime(2025, 1, 2, 0, 0)
    result = _get_python_value(value, "utf8")
    assert result == expected


def test_get_python_value_datetime():
    value = "2025-01-02 12:24:36"
    expected = datetime.datetime(2025, 1, 2, 12, 24, 36)
    result = _get_python_value(value, "utf8")
    assert result == expected


def test_get_python_value_duration():
    value = "PT12H24M36S"
    expected = datetime.timedelta(hours=12, minutes=24, seconds=36)
    result = _get_python_value(value, "utf8")
    assert result == expected


def test_table_protection_key():
    table = Table("name", protected=True, protection_key="hash")
    assert table.protected
    assert table.protection_key == "hash"


def test_table_protection_key_raise():
    with pytest.raises(ValueError):
        Table("name", protected=True)


def test_table_translate_table_coordinates_list_1(table):
    result = table._translate_table_coordinates_list([4])
    assert result == (None, 4, None, 4)


def test_table_translate_table_coordinates_list_2(table):
    result = table._translate_table_coordinates_list([-2])
    assert result == (None, 2, None, 2)


def test_table_translate_table_coordinates_list_3(table):
    result = table._translate_table_coordinates_list([1, 2])
    assert result == (None, 1, None, 2)


def test_table_translate_table_coordinates_list_4(table):
    result = table._translate_table_coordinates_list([-2, 2])
    assert result == (None, 2, None, 2)


def test_table_translate_table_coordinates_list_5(table):
    result = table._translate_table_coordinates_list([-1, 2])
    assert result == (None, 3, None, 2)


def test_table_translate_table_coordinates_list_6(table):
    result = table._translate_table_coordinates_list([-2, -1])
    assert result == (None, 2, None, 3)


def test_table_translate_table_coordinates_list_7(table):
    result = table._translate_table_coordinates_list([1, 1, 1, 1])
    assert result == (1, 1, 1, 1)


def test_table_translate_table_coordinates_list_8(table):
    result = table._translate_table_coordinates_list([-1, 1, 1, 1])
    assert result == (6, 1, 1, 1)


def test_table_translate_table_coordinates_list_9(table):
    result = table._translate_table_coordinates_list([1, -1, 1, 1])
    assert result == (1, 3, 1, 1)


def test_table_translate_table_coordinates_list_10(table):
    result = table._translate_table_coordinates_list([1, 1, -1, 1])
    assert result == (1, 1, 6, 1)


def test_table_translate_table_coordinates_list_11(table):
    result = table._translate_table_coordinates_list([1, 1, 1, -1])
    assert result == (1, 1, 1, 3)


def test_table_translate_table_coordinates_list_12(table):
    result = table._translate_table_coordinates_list([-1, -1, -1, -1])
    assert result == (6, 3, 6, 3)


def test_table_translate_table_coordinates_str_1(table):
    result = table._translate_table_coordinates_str("5")
    assert result == (None, 4, None, 4)


def test_table_translate_table_coordinates_str_1b(table):
    result = table._translate_table_coordinates_str("5:5")
    assert result == (None, 4, None, 4)


def test_table_translate_table_coordinates_str_2(table):
    result = table._translate_table_coordinates_str("1:1")
    assert result == (None, 0, None, 0)


def test_table_translate_table_coordinates_str_3(table):
    result = table._translate_table_coordinates_str("2:3")
    assert result == (None, 1, None, 2)


def test_table_translate_table_coordinates_str_4(table):
    result = table._translate_table_coordinates_str("3:3")
    assert result == (None, 2, None, 2)


def test_table_translate_table_coordinates_str_5(table):
    result = table._translate_table_coordinates_str("4:3")
    assert result == (None, 3, None, 2)


def test_table_translate_table_coordinates_str_7(table):
    result = table._translate_table_coordinates_str("B2:B2")
    assert result == (1, 1, 1, 1)


def test_table_translate_table_coordinates_str_8(table):
    result = table._translate_table_coordinates_str("G2:B2")
    assert result == (6, 1, 1, 1)


def test_table_translate_table_coordinates_str_9(table):
    result = table._translate_table_coordinates_str("B4:B2")
    assert result == (1, 3, 1, 1)


def test_table_translate_table_coordinates_str_10(table):
    result = table._translate_table_coordinates_str("B2:G2")
    assert result == (1, 1, 6, 1)


def test_table_translate_table_coordinates_str_11(table):
    result = table._translate_table_coordinates_str("B2:B4")
    assert result == (1, 1, 1, 3)


def test_table_translate_table_coordinates_str_12(table):
    result = table._translate_table_coordinates_str("G4:G4")
    assert result == (6, 3, 6, 3)


def test_table_translate_column_coordinates_list_1(table):
    result = table._translate_column_coordinates_list([0])
    assert result == (0, None, 0, None)


def test_table_translate_column_coordinates_list_2(table):
    result = table._translate_column_coordinates_list([4])
    assert result == (4, None, 4, None)


def test_table_translate_column_coordinates_list_3(table):
    result = table._translate_column_coordinates_list([-2])
    assert result == (5, None, 5, None)


def test_table_translate_column_coordinates_list_4(table):
    result = table._translate_column_coordinates_list([0, 0])
    assert result == (0, None, 0, None)


def test_table_translate_column_coordinates_list_5(table):
    result = table._translate_column_coordinates_list([2, 3])
    assert result == (2, None, 3, None)


def test_table_translate_column_coordinates_list_6(table):
    result = table._translate_column_coordinates_list([-2, 3])
    assert result == (5, None, 3, None)


def test_table_translate_column_coordinates_list_7(table):
    result = table._translate_column_coordinates_list([-1, -2])
    assert result == (6, None, 5, None)


def test_table_translate_column_coordinates_list_8(table):
    result = table._translate_column_coordinates_list([0, 0, 0, 0])
    assert result == (0, 0, 0, 0)


def test_table_translate_column_coordinates_list_9(table):
    result = table._translate_column_coordinates_list([1, 2, 3, 4])
    assert result == (1, 2, 3, 4)


def test_table_translate_column_coordinates_list_10(table):
    result = table._translate_column_coordinates_list([-1, -2, -3, -4])
    assert result == (6, 2, 4, 0)


def test_table_translate_column_coordinates_str_1(table):
    result = table._translate_column_coordinates_str("1")
    assert result == (None, 0, None, 0)


def test_table_translate_column_coordinates_str_2(table):
    result = table._translate_column_coordinates_str("A1")
    assert result == (0, 0, 0, 0)


def test_table_translate_column_coordinates_str_3(table):
    result = table._translate_column_coordinates_str("A1:B3")
    assert result == (0, 0, 1, 2)


def test_table_translate_column_coordinates_str_4(table):
    result = table._translate_column_coordinates_str("A:A")
    assert result == (0, None, 0, None)


def test_table_translate_column_coordinates_str_5(table):
    result = table._translate_column_coordinates_str("A:B")
    assert result == (0, None, 1, None)


def test_table_translate_column_coordinates_str_6(table):
    result = table._translate_column_coordinates_str("B")
    assert result == (1, None, 1, None)


def test_table_translate_column_coordinates_1(table):
    result = table._translate_column_coordinates("B")
    assert result == (1, None, 1, None)


def test_table_translate_column_coordinates_2(table):
    result = table._translate_column_coordinates([1])
    assert result == (1, None, 1, None)


def test_translate_cell_coordinates_1(table):
    result = table._translate_cell_coordinates([1, 2])
    assert result == (1, 2)


def test_translate_cell_coordinates_2(table):
    result = table._translate_cell_coordinates([1, 2, 3, 4])
    assert result == (1, 2)


def test_translate_cell_coordinates_3(table):
    with pytest.raises(ValueError):
        table._translate_cell_coordinates([1, 2, 3])


def test_translate_cell_coordinates_4(table):
    result = table._translate_cell_coordinates("A1:d6")
    assert result == (0, 0)


def test_translate_cell_coordinates_5(table):
    result = table._translate_cell_coordinates("B2")
    assert result == (1, 1)


def test_translate_cell_coordinates_6(table):
    result = table._translate_cell_coordinates([-5, -5])
    assert result == (2, 3)


def test_table_get_formatted_text_normal_1(table):
    result = table._get_formatted_text_normal(None)
    expected = (
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n2\n3\n4\n5\n6\n7\n\n"
    )
    assert result == expected


def test_table_get_formatted_text_normal_2(table):
    table.set_value((1, 1), None)
    result = table._get_formatted_text_normal(None)
    expected = (
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n2\n3\n4\n5\n6\n7\n\n"
    )
    assert result == expected


def test_table_get_formatted_text_normal_3(table):
    table.set_value((1, 1), None)
    cell = table.get_cell((1, 1))
    item = Header(1, "test")
    cell.append(item)
    table.set_cell((1, 1), cell)
    result = table._get_formatted_text_normal(None)
    expected = (
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\ntest\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n2\n3\n4\n5\n6\n7\n\n"
    )
    assert result == expected


def test_table_get_formatted_text_rst_1(table):
    ctx = {"no_img_level": 0}
    result = table._get_formatted_text_rst(ctx)
    expected = (
        "\n"
        "== == == == == == == \n"
        "1  1  1  2  3  3  3  \n"
        "1  1  1  2  3  3  3  \n"
        "1  1  1  2  3  3  3  \n"
        "1  2  3  4  5  6  7  \n"
        "== == == == == == == \n\n"
    )
    assert result == expected


def test_table_get_formatted_text_rst_2(table):
    table.set_value((1, 1), None)
    cell = table.get_cell((1, 1))
    item = Header(1, "test")
    cell.append(item)
    table.set_cell((1, 1), cell)
    ctx = {
        "no_img_level": 0,
        "rst_mode": True,
    }
    result = table._get_formatted_text_rst(ctx)
    expected = (
        "\n"
        "== ========= == == == == == \n"
        "1  1         1  2  3  3  3  \n"
        "1  test      1  2  3  3  3  \n"
        "   ####                     \n"
        "1  1         1  2  3  3  3  \n"
        "1  2         3  4  5  6  7  \n"
        "== ========= == == == == == \n\n"
    )
    assert result == expected


def test_table_get_formatted_text_rst_3(table):
    table.set_value((1, 1), None)
    cell = table.get_cell((1, 1))
    item = Header(1, "test")
    cell.append(item)
    table.set_cell((1, 1), cell)
    ctx = {
        "no_img_level": 0,
        "rst_mode": True,
    }
    table.set_value((1, 2), "")
    result = table._get_formatted_text_rst(ctx)
    expected = (
        "\n== ========= == == == == == \n"
        "1  1         1  2  3  3  3  \n"
        "1  test      1  2  3  3  3  \n"
        "   ####                     \n"
        "1            1  2  3  3  3  \n"
        "1  2         3  4  5  6  7  \n"
        "== ========= == == == == == \n\n"
    )
    assert result == expected


def test_table_get_formatted_text_rst_empty():
    table = Table("empty")
    ctx = {
        "no_img_level": 0,
        "rst_mode": True,
    }
    result = table._get_formatted_text_rst(ctx)
    expected = ""
    assert result == expected


def test_table_get_formatted_text_rst_empty_2():
    table = Table("empty", 1, 1)
    ctx = {
        "no_img_level": 0,
        "rst_mode": True,
    }
    result = table._get_formatted_text_rst(ctx)
    expected = ""
    assert result == expected


def test_table_printable_1(table):
    result = table.printable
    expected = False
    assert result is expected


def test_table_printable_2():
    table = table = Table("empty", 1, 1)
    result = table.printable
    expected = True
    assert result is expected


def test_table_printable_3(table):
    table.printable = True
    result = table.printable
    expected = True
    assert result is expected


def test_table_printable_4(table):
    table.printable = False
    result = table.printable
    expected = False
    assert result is expected


def test_table_ranges_1(table):
    result = table.print_ranges
    expected = []
    assert result == expected


def test_table_ranges_2(table):
    table.print_ranges = ["A1:B2", "A4:C4"]
    result = table.print_ranges
    expected = ["A1:B2", "A4:C4"]
    assert result == expected


def test_table_get_formatted_text_1(table):
    result = table.get_formatted_text(None)
    expected = (
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n1\n1\n2\n3\n3\n3\n\n"
        "1\n2\n3\n4\n5\n6\n7\n\n"
    )
    assert result == expected


def test_table_get_formatted_text_2(table):
    ctx = {
        "no_img_level": 0,
        "rst_mode": True,
    }
    result = table.get_formatted_text(ctx)
    expected = (
        "\n"
        "== == == == == == == \n"
        "1  1  1  2  3  3  3  \n"
        "1  1  1  2  3  3  3  \n"
        "1  1  1  2  3  3  3  \n"
        "1  2  3  4  5  6  7  \n"
        "== == == == == == == \n\n"
    )
    assert result == expected


def test_table_append_column(table):
    col = Column()
    table.append(col)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 1, 1, 2, 3, 3, 3, None],
        [1, 2, 3, 4, 5, 6, 7, None],
    ]


def test_table_append_row(table):
    row = Row()
    table.append(row)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
    ]


def test_table_append_somthing(table):
    elem = Element.from_tag("text:p")
    table.append(elem)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_iter_values_1(table):
    result = table.iter_values()
    assert isinstance(result, Iterator)


def test_table_iter_values_2(table):
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_iter_values_3(table):
    coord = (2, 2, 3, 3)
    result = list(table.iter_values(coord))
    assert result == [
        [1, 2],
        [3, 4],
    ]


def test_table_iter_values_4(table):
    table.set_value((2, 2), None)
    coord = (2, 2, 3, 3)
    result = list(table.iter_values(coord))
    assert result == [
        [None, 2],
        [3, 4],
    ]


def test_table_iter_values_5(table):
    row = Row()
    row.set_values([10, 20, 30])
    table.append(row)
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [10, 20, 30, None, None, None, None],
    ]


def test_table_iter_values_6(table):
    row = Row()
    row.set_values([10, 20, 30])
    table.append(row)
    result = list(table.iter_values(complete=True, get_type=True))
    assert result == [
        [
            (1, "float"),
            (1, "float"),
            (1, "float"),
            (2, "float"),
            (3, "float"),
            (3, "float"),
            (3, "float"),
        ],
        [
            (1, "float"),
            (1, "float"),
            (1, "float"),
            (2, "float"),
            (3, "float"),
            (3, "float"),
            (3, "float"),
        ],
        [
            (1, "float"),
            (1, "float"),
            (1, "float"),
            (2, "float"),
            (3, "float"),
            (3, "float"),
            (3, "float"),
        ],
        [
            (1, "float"),
            (2, "float"),
            (3, "float"),
            (4, "float"),
            (5, "float"),
            (6, "float"),
            (7, "float"),
        ],
        [
            (10, "float"),
            (20, "float"),
            (30, "float"),
            (None, None),
            (None, None),
            (None, None),
            (None, None),
        ],
    ]


def test_table_iter_values_7(table):
    row = Row()
    row.set_values([10, 20, 30])
    table.append(row)
    result = list(table.iter_values(complete=False))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [10, 20, 30],
    ]


def test_table_set_values_1(table):
    table.set_values(values=[[10, 20, 30]], coord=(None, None))
    result = list(table.iter_values())
    assert result == [
        [10, 20, 30, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_values_2(table):
    table.rows[3].repeated = 4
    table.set_values(values=[[], [10, 20, 30]], coord=(1, 2))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 10, 20, 30, 5, 6, 7],
        [1, 2, 3, 4, 5, 6, 7],
        [1, 2, 3, 4, 5, 6, 7],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_is_empty(table):
    assert not table.is_empty()


def test_table_extend_rows_1(table):
    row1 = Row()
    row2 = Row()
    row2.set_values([10, 20, 30, 40, 50])
    table.extend_rows([row1, row2])
    result = list(table.iter_values(complete=True))
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
        [10, 20, 30, 40, 50, None, None],
    ]


def test_table_extend_rows_2(table):
    table.extend_rows([])
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_extend_rows_3(table):
    table.extend_rows()
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_extend_rows_4(table):
    row = Row()
    row.set_values([10, 20, 30, 40, 50, 60, 70, 80, 90])
    table.extend_rows([row])
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
        [10, 20, 30, 40, 50, 60, 70, 80, 90],
    ]


def test_table_delete_row(table):
    table.delete_row(99)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_get_row_values_1(table):
    result = table.get_row_values(1, get_type=True)
    assert result == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
    ]


def test_table_get_row_values_2(table):
    result = table.get_row_values(1, complete=False)
    assert result == [1, 1, 1, 2, 3, 3, 3]


def test_table_set_row_cells_1(table):
    c1 = Cell(value=40)
    c2 = Cell(value=50)
    table.set_row_cells(2, [c1, c2])
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [40, 50, None, None, None, None, None],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_row_cells_2(table):
    table.set_row_cells(2, [])
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [None, None, None, None, None, None, None],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_row_cells_3(table):
    table.set_row_cells(2)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [None, None, None, None, None, None, None],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_insert_row_empty_1(table):
    table.insert_row(0, row=None, clone=False)
    result = list(table.iter_values())
    assert result == [
        [None, None, None, None, None, None, None],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_insert_row_empty_2(table):
    table.insert_row(3, row=None, clone=False)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [None, None, None, None, None, None, None],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_insert_row_empty_3(table):
    table.insert_row(4, row=None, clone=False)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
    ]


def test_table_insert_row_empty_4(table):
    table.insert_row(5, row=None, clone=False)
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None],
    ]


def test_table_get_cell_raise_1(table):
    with pytest.raises(ValueError):
        table.get_cell((None, 0))


def test_table_get_cell_raise_2(table):
    with pytest.raises(ValueError):
        table.get_cell((0, None))


def test_table_get_value_raise_1(table):
    with pytest.raises(ValueError):
        table.get_value((None, 0))


def test_table_get_value_raise_2(table):
    with pytest.raises(ValueError):
        table.get_value((0, None))


def test_table_get_value_none_1(table):
    result = table.get_value((0, 99), get_type=False)
    assert result is None


def test_table_get_value_none_2(table):
    result = table.get_value((0, 99), get_type=True)
    assert result == (None, None)


def test_table_get_value_none_3(table):
    result = table.get_value((99, 2), get_type=False)
    assert result is None


def test_table_get_value_none_4(table):
    result = table.get_value((99, 2), get_type=True)
    assert result == (None, None)


def test_table_set_cell_raise_1(table):
    with pytest.raises(ValueError):
        table.set_cell((None, 0))


def test_table_set_cell_raise_2(table):
    with pytest.raises(ValueError):
        table.set_cell((0, None))


def test_table_insert_cell_raise_1(table):
    with pytest.raises(ValueError):
        table.insert_cell((None, 0))


def test_table_insert_cell_raise_2(table):
    with pytest.raises(ValueError):
        table.insert_cell((0, None))


def test_table_delete_cell_raise_1(table):
    with pytest.raises(ValueError):
        table.delete_cell((None, 0))


def test_table_delete_cell_raise_2(table):
    with pytest.raises(ValueError):
        table.delete_cell((0, None))


def test_table_set_cells_none_1(table):
    table.set_cells([])
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_cells_none_2(table):
    table.set_cells([], (None, 0))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_cells_none_3(table):
    table.set_cells([], (0, None))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_cells_1(table):
    r1 = [Cell(10), Cell(11), Cell(12)]
    table.set_cells([r1], (1, 1))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 10, 11, 12, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]


def test_table_set_cells_2(table):
    r1 = [Cell(10), Cell(11), Cell(12)]
    r3 = [Cell(30), Cell(31), Cell(32)]
    table.set_cells([r1, None, r3], (1, 1))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 10, 11, 12, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 30, 31, 32, 5, 6, 7],
    ]


def test_table_set_cells_3(table):
    r1 = [Cell(10), Cell(11), Cell(12)]
    table.set_cells([r1], (1, 8))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None],
        [None, 10, 11, 12, None, None, None],
    ]


def test_table_set_cells_4(table):
    r1 = [Cell(10), Cell(11), Cell(12)]
    table.set_cells([r1], (1, 8))
    r2 = [Cell(30), Cell(31), Cell(32)]
    table.set_cells([r2], (1, 6))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None],
        [None, 30, 31, 32, None, None, None],
        [None, None, None, None, None, None, None],
        [None, 10, 11, 12, None, None, None],
    ]


def test_table_delete_cell_big_y(table):
    table.delete_cell((0, 99))
    result = list(table.iter_values())
    assert result == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest

from odfdo.scripts import to_markdown
from odfdo.scripts.to_markdown import main as main_script
from odfdo.scripts.to_markdown import main_to_md, parse_cli_args

SCRIPT = Path(to_markdown.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_to_md_no_param():
    params = []
    out, _err, exitcode = run_params(params)
    assert exitcode == 1
    assert "usage:" in out


# direct access to internal function


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Fails on Windows due to stdin/output capture conflict.",
)
def test_to_md_2_no_param_on_main_function(monkeypatch):
    with pytest.raises(ValueError) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.value.code >= 1


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Fails on Windows due to stdin/output capture conflict.",
)
def test_to_md_2_no_param():
    with pytest.raises(ValueError) as result:
        params = parse_cli_args([])
        main_to_md(params)
        assert result.value.code >= 1


def test_to_md_2_version(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--version"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "odfdo-markdown v3" in captured.out


def test_to_md_2_help(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--help"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "document to Markdown" in captured.out


def test_to_md_2_no_file():
    params = parse_cli_args(["none_file"])
    with pytest.raises(FileNotFoundError) as result:
        main_to_md(params)
        assert result.value.code >= 1


def test_to_md_2_md_osq_bad(samples):
    source = samples("simple_table.ods")
    params = parse_cli_args([str(source)])

    with pytest.raises(NotImplementedError) as result:
        main_to_md(params)
        assert result.value.code >= 1


def test_to_md_2_example(capsys, samples):
    source = samples("example.odt")
    params = parse_cli_args([str(source)])

    main_to_md(params)
    captured = capsys.readouterr()

    assert "odfdo Test Case Document" in captured.out
    assert "First paragraph" in captured.out



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2012 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Row class for "table:table-row" tag."""

from __future__ import annotations

import contextlib
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any, cast

from .cell import Cell
from .element import (
    Element,
    register_element_class,
    xpath_compile,
    xpath_return_elements,
)
from .table_cache import _XP_CELL_IDX, RowCache, TableCache
from .utils import convert_coordinates, increment, translate_from_any

if TYPE_CHECKING:
    from lxml.etree import XPath

    from .style import Style


_XPATH_CELL = xpath_compile("(table:table-cell|table:covered-table-cell)")


class Row(Element):
    """A row of a table, "table:table-row".

    Args:
        width (int, optional): The number of cells to create in the row.
        repeated (int, optional): The number of times the row is repeated.
        style (str, optional): The style name for the row.
    """

    _tag = "table:table-row"
    _append = Element.append

    def __init__(
        self,
        width: int | None = None,
        repeated: int | None = None,
        style: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a Row, "table:table-row", optionally filled with "width"
        number of cells.

        Rows contain cells, their number determine the number of columns.

        You don't generally have to create rows by hand, use the Table API.

        Args:
            width: The number of cells to create in the row.
            repeated: The number of times the row is repeated.
            style: The style name for the row.
        """
        super().__init__(**kwargs)
        self._table_cache = TableCache()
        self._row_cache = RowCache()
        self.y = None
        self._compute_row_cache()
        if self._do_init:
            if width is not None:
                for _i in range(width):
                    self.append(Cell())
            if repeated:
                self.repeated = repeated
            if style is not None:
                self.style = style
            self._compute_row_cache()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} y={self.y}>"

    def get_elements(self, xpath_query: XPath | str) -> list[Element]:
        """Get a list of elements matching the XPath query.

        Args:
            xpath_query: The XPath query.

        Returns:
            list[Element]: A list of matching elements.
        """
        if isinstance(xpath_query, str):
            elements = xpath_return_elements(
                xpath_compile(xpath_query),
                self._xml_element,
            )
        else:
            elements = xpath_return_elements(
                xpath_query,
                self._xml_element,
            )
        cache = (self._table_cache, self._row_cache)
        return [Element.from_tag_for_clone(e, cache) for e in elements]

    def _copy_cache(self, cache: tuple) -> None:
        """Copy cache when cloning.

        Args:
            cache: The cache to copy.
        """
        self._table_cache = cache[0]
        if cache[1]:  # pragma: no cover
            self._row_cache = cache[1]

    def clear(self) -> None:
        """Remove text, children and attributes from the Row."""
        self._xml_element.clear()
        self._table_cache = TableCache()
        self._row_cache = RowCache()

    def _get_cells(self) -> list[Cell]:
        return cast(list[Cell], self.get_elements(_XPATH_CELL))

    def _translate_row_coordinates(
        self,
        coord: tuple | list | str,
    ) -> tuple[int | None, int | None]:
        xyzt = convert_coordinates(coord)
        if len(xyzt) == 2:
            x, z = xyzt
        else:
            x, _, z, __ = xyzt
        if x and x < 0:
            x = increment(x, self.width)
        if z and z < 0:
            z = increment(z, self.width)
        return (x, z)

    def _compute_row_cache(self) -> None:
        idx_repeated_seq = self.elements_repeated_sequence(
            _XPATH_CELL, "table:number-columns-repeated"
        )
        self._row_cache.make_cell_map(idx_repeated_seq)

    # Public API

    @property
    def clone(self) -> Row:
        """Return a copy of the row."""
        cloned_row: Row = Element.clone.fget(self)  # type: ignore[attr-defined]
        cloned_row.y = self.y
        cloned_row._table_cache = TableCache.copy(self._table_cache)
        cloned_row._row_cache = RowCache.copy(self._row_cache)
        return cloned_row

    def _set_repeated(self, repeated: int | None) -> None:
        """Set the number of times the row is repeated.

        Internal use only. Without changing cache.

        Args:
            repeated: The number of repetitions.
        """
        if repeated is None or repeated < 2:
            with contextlib.suppress(KeyError):
                self.del_attribute("table:number-rows-repeated")
            return
        self.set_attribute("table:number-rows-repeated", str(repeated))

    @property
    def repeated(self) -> int | None:
        """Get / set the number of times the row is repeated.

        Always None when using the table API.

        Returns:
            int | None: The number of repetitions.
        """
        repeated = self.get_attribute("table:number-rows-repeated")
        if repeated is None:
            return None
        return int(repeated)

    @repeated.setter
    def repeated(self, repeated: int | None) -> None:
        self._set_repeated(repeated)
        # update cache
        current: Element = self
        while True:
            # look for Table, parent may be group of rows
            upper = current.parent
            if not upper:
                # lonely row
                return
            # parent may be group of rows, not table
            if isinstance(upper, Element) and upper._tag == "table:table":
                upper._table_cache = self._table_cache  # type: ignore[attr-defined]
                upper._compute_table_cache()  # type: ignore[attr-defined]
                return
            current = upper

    @property
    def style(self) -> str | None:
        """Get /set the style of the row itself.

        Returns:
            str | None: The style name.
        """
        return self.get_attribute_string("table:style-name")

    @style.setter
    def style(self, style: str | Style) -> None:
        self.set_style_attribute("table:style-name", style)

    @property
    def width(self) -> int:
        """Get the number of expected cells in the row, including repetitions.

        Returns:
            int: The width of the row.
        """
        return self._row_cache.width()

    def _translate_x_from_any(self, x: str | int) -> int:
        return translate_from_any(x, self.width, 0)

    def _yield_odf_cells(self) -> Iterator[Cell]:
        for cell in self._get_cells():
            if cell.repeated is None:
                yield cell
            else:
                for _ in range(cell.repeated):
                    cell_copy = cell.clone
                    cell_copy.repeated = None
                    yield cell_copy

    def iter_cells(
        self,
        start: int | None = None,
        end: int | None = None,
    ) -> Iterator[Cell]:
        """Yields Cell elements, expanding repetitions.

        This method produces individual Cell objects. The yielded
        Cell are copies; use `set_cell()` to apply changes.

        Args:
            start: The starting cell index (0-based). Defaults to 0.
            end: The ending cell index (inclusive). Defaults to 2**32.

        Yields:
            Cell: The next Cell element in the specified range..
        """
        if start is None:
            start = 0
        start = max(0, start)
        if end is None:
            end = 2**32
        if end < start:
            return
        x: int = -1
        for cell in self._yield_odf_cells():
            x += 1
            if x < start:
                continue
            if x > end:
                return
            cell.x = x
            cell.y = self.y
            yield cell

    traverse = iter_cells

    def get_cells(
        self,
        coord: str | tuple | None = None,
        style: str | None = None,
        content: str | None = None,
        cell_type: str | None = None,
    ) -> list[Cell]:
        """Get the list of cells matching the criteria.

        Filter by cell_type, with cell_type 'all' will retrieve cells of any
        type, aka non empty cells.

        Filter by coordinates will retrieve the amount of cells defined by
        'coord', minus the other filters.

        Args:
            coord: The coordinates of the cells.
            style: The style name to filter by.
            content: A regex to match in the cell content.
            cell_type: The type of the cell. Can be 'boolean',
                'float', 'date', 'string', 'time', 'currency', 'percentage'
                or 'all'.

        Returns:
            list[Cell]: A list of matching cells.
        """
        # fixme : not clones ?
        if coord:
            x, z = self._translate_row_coordinates(coord)
        else:
            x = None
            z = None
        if cell_type:
            cell_type = cell_type.lower().strip()
        cells: list[Cell] = []
        for cell in self.iter_cells(start=x, end=z):
            # Filter the cells by cell_type
            if cell_type:
                ctype = cell.type
                if not ctype or not (ctype == cell_type or cell_type == "all"):
                    continue
            # Filter the cells with the regex
            if content and not cell.match(content):
                continue
            # Filter the cells with the style
            if style and style != cell.style:
                continue
            cells.append(cell)
        return cells

    @property
    def cells(self) -> list[Cell]:
        """Get the list of all cells.

        Returns:
            A list of all cells.
        """
        # fixme : not clones ?
        return list(self.iter_cells())

    def _get_cell2(self, x: int, clone: bool = True) -> Cell | None:
        if x >= self.width:
            return Cell()
        base_cell = self._get_cell2_base(x)
        if base_cell is None:  # pragma: no cover
            return None
        if clone:
            return base_cell.clone
        return base_cell

    def _get_cell2_base(self, x: int) -> Cell | None:
        idx = self._row_cache.cell_idx(x)
        if idx is None:
            return None
        cell: Cell | None = self._row_cache.cached_cell(idx)
        if cell is None:
            cell = self._get_element_idx2(_XP_CELL_IDX, idx)  # type: ignore[assignment]
            if cell is None:  # pragma: no cover
                return None
            self._row_cache.store_cell(cell, idx)
        return cell

    def get_cell(self, x: int, clone: bool = True) -> Cell | None:
        """Get the cell at position "x". Alphabetical positions like "D" are
        also accepted.

        A copy is returned, so changes will not affect the document.
        Use `set_cell` to apply changes.

        Args:
            x: The column index or name.
            clone: Whether to return a copy of the cell.

        Returns:
            Cell | None: The cell at the given position.
        """
        x = self._translate_x_from_any(x)
        cell = self._get_cell2(x, clone=clone)
        if not cell:
            return None  # pragma: no cover
        cell.y = self.y
        cell.x = x
        return cell

    def get_value(
        self,
        x: int | str,
        get_type: bool = False,
    ) -> Any | tuple[Any, str]:
        """Shortcut to get the value of the cell at position "x".

        If `get_type` is True, returns a tuple (value, ODF type).
        If the cell is empty, returns None or (None, None).

        Args:
            x: The column index or name.
            get_type: Whether to return the ODF type of the value.

        Returns:
            Any | tuple[Any, str]: The value of the cell, or a tuple (value, type).
        """
        if get_type:
            x = self._translate_x_from_any(x)
            cell = self._get_cell2_base(x)
            if cell is None:
                return (None, None)
            return cell.get_value(get_type=get_type)
        x = self._translate_x_from_any(x)
        cell = self._get_cell2_base(x)
        if cell is None:
            return None
        return cell.get_value()

    def set_cell(
        self,
        x: int | str,
        cell: Cell | None = None,
        clone: bool = True,
    ) -> Cell:
        """Set the cell at position "x". Alphabetical positions like "D" are
        also accepted.

        Args:
            x: The column index or name.
            cell: The cell to set. If None, an empty cell is
                created.
            clone: Whether to clone the cell before setting it.

        Returns:
            Cell: The cell that was set.
        """
        cell_back: Cell
        if cell is None:
            cell = Cell()
            repeated = 1
            clone = False
        else:
            repeated = cell.repeated or 1
        x = self._translate_x_from_any(x)
        # Outside the defined row
        diff = x - self.width
        if diff == 0:
            cell_back = self.append_cell(cell, _repeated=repeated, clone=clone)
        elif diff > 0:
            self.append_cell(Cell(repeated=diff), _repeated=diff, clone=False)
            cell_back = self.append_cell(cell, _repeated=repeated, clone=clone)
        else:
            # Inside the defined row
            cell_back = self._row_cache.set_cell_in_cache(x, cell, self, clone=clone)
            cell_back.x = x
            cell_back.y = self.y
        return cell_back

    def set_value(
        self,
        x: int | str,
        value: Any,
        style: str | None = None,
        cell_type: str | None = None,
        currency: str | None = None,
    ) -> None:
        """Shortcut to set the value of the cell at position "x".

        Args:
            x: The column index or name.
            value: The value to set.
            style: The style to apply to the cell.
            cell_type: The type of the cell. Can be 'boolean',
                'currency', 'date', 'float', 'string', 'time', 'currency' or
                'percentage'.
            currency: The currency symbol if the type is
                'currency'.
        """
        self.set_cell(
            x,
            Cell(value, style=style, cell_type=cell_type, currency=currency),
            clone=False,
        )

    def insert_cell(
        self,
        x: int | str,
        cell: Cell | None = None,
        clone: bool = True,
    ) -> Cell:
        """Insert a cell at position "x". If no cell is given, an empty one is
        created.

        Alphabetical positions like "D" are also accepted.

        Do not use when working on a table, use Table.insert_cell().

        Args:
            x: The column index or name.
            cell: The cell to insert.
            clone: Whether to clone the cell before inserting it.

        Returns:
            Cell: The cell that was inserted.
        """
        cell_back: Cell
        if cell is None:
            cell = Cell()
        x = self._translate_x_from_any(x)
        # Outside the defined row
        diff = x - self.width
        if diff < 0:
            cell_back = self._row_cache.insert_cell_in_cache(x, cell, self)
            cell_back.x = x
            cell_back.y = self.y
        elif diff == 0:
            cell_back = self.append_cell(cell, clone=clone)
        else:
            self.append_cell(Cell(repeated=diff), _repeated=diff, clone=False)
            cell_back = self.append_cell(cell, clone=clone)
        return cell_back

    def extend_cells(self, cells: Iterable[Cell] | None = None) -> None:
        """Extend the row with a list of cells.

        Args:
            cells: The cells to append.
        """
        if cells is None:
            cells = []
        self.extend(cells)
        self._compute_row_cache()

    def append_cell(
        self,
        cell: Cell | None = None,
        clone: bool = True,
        _repeated: int | None = None,
    ) -> Cell:
        """Append a cell at the end of the row. Repeated cells are accepted.
        If no cell is given, an empty one is created.

        Do not use when working on a table, use Table.append_cell().

        Args:
            cell: The cell to append.
            clone: Whether to clone the cell before appending it.
            _repeated: The repeated value of the cell.

        Returns:
            Cell: The cell that was appended.
        """
        if cell is None:
            cell = Cell()
            clone = False
        if clone:
            cell = cell.clone
        self._append(cell)
        if _repeated is None:
            _repeated = cell.repeated or 1
        self._row_cache.insert_cell_map_once(_repeated)
        cell.x = self.width - 1
        cell.y = self.y
        return cell

    # fix for unit test and typos
    append = append_cell  # type:ignore[assignment]

    def delete_cell(self, x: int | str) -> None:
        """Delete the cell at the given position "x". Alphabetical positions
        like "D" are also accepted.

        Cells on the right will be shifted to the left. In a table, other
        rows remain unaffected.

        Args:
            x: The column index or name.
        """
        x = self._translate_x_from_any(x)
        if x >= self.width:
            return
        self._row_cache.delete_cell_in_cache(x, self)

    def get_values(
        self,
        coord: str | tuple | None = None,
        cell_type: str | None = None,
        complete: bool = False,
        get_type: bool = False,
    ) -> list[Any | tuple[Any, Any]]:
        """Shortcut to get the cell values in this row.

        - Filter by `cell_type`: with 'all' will retrieve cells of any type
          (non-empty).
        - If `cell_type` is used and `complete` is True, missing values are
          replaced by None.
        - If `cell_type` is None, `complete` is always True: `get_values()`
          returns None for each empty cell, the length of the list is equal
          to the length of the row (depending on coordinates use).
        - If `get_type` is True, returns a tuple (value, ODF type of value),
          or (None, None) for empty cells if `complete` is True.
        - Filter by `coord` will retrieve the amount of cells defined by
          coordinates with None for empty cells, except when using `cell_type`.

        Args:
            coord: Coordinates in the row.
            cell_type: Type of cell to filter by. Can be
                'boolean', 'float', 'date', 'string', 'time', 'currency',
                'percentage' or 'all'.
            complete: Whether to include empty cells as None.
            get_type: Whether to return the ODF type of the value.

        Returns:
            list[Any | tuple[Any, Any]]: A list of values, or a list of (value, type) tuples.
        """
        if coord:
            x, z = self._translate_row_coordinates(coord)
        else:
            x = None
            z = None
        if cell_type:
            cell_type = cell_type.lower().strip()
            values: list[Any | tuple[Any, Any]] = []
            for cell in self.iter_cells(start=x, end=z):
                # Filter the cells by cell_type
                ctype = cell.type
                if not ctype or not (ctype == cell_type or cell_type == "all"):
                    if complete:
                        if get_type:
                            values.append((None, None))
                        else:
                            values.append(None)
                    continue
                values.append(cell.get_value(get_type=get_type))
            return values
        else:
            return [
                cell.get_value(get_type=get_type)
                for cell in self.iter_cells(start=x, end=z)
            ]

    def get_sub_elements(
        self,
    ) -> list[Any]:
        """Shortcut to get the Elements inside cells in this row.

        Missing values are replaced by None. Cell type should be always
        'string' when using this method, the length of the list is equal
        to the length of the row.

        Returns:
            list[Any]: A list of elements.
        """
        return [cell.children for cell in self.iter_cells()]

    def set_cells(
        self,
        cells: list[Cell] | tuple[Cell] | None = None,
        start: int | str = 0,
        clone: bool = True,
    ) -> None:
        """Set the cells in the row, from the 'start' column. This method does
        not clear the row, use row.clear() before to start with an empty row.

        Args:
            cells: The cells to set.
            start: The starting column index or name.
            clone: Whether to clone the cells before setting them.
        """
        if cells is None:
            cells = []
        if start is None:
            start = 0
        else:
            start = self._translate_x_from_any(start)
        if start == 0 and clone is False and (len(cells) >= self.width):
            self.clear()
            self.extend_cells(cells)
        else:
            x = start
            for cell in cells:
                self.set_cell(x, cell, clone=clone)
                if cell:
                    x += cell.repeated or 1
                else:
                    x += 1

    def set_values(
        self,
        values: list[Any],
        start: int | str = 0,
        style: str | None = None,
        cell_type: str | None = None,
        currency: str | None = None,
    ) -> None:
        """Shortcut to set the value of cells in the row, from the 'start'
        column with values. This method does not clear the row, use row.clear()
        before to start with an empty row.

        Args:
            values: A list of values to set.
            start: The starting column index or name.
            style: The style to apply to the cells.
            cell_type: The type of the cells. Can be
                'boolean', 'float', 'date', 'string', 'time', 'currency' or
                'percentage'.
            currency: The currency symbol if the type is
                'currency'.
        """
        # fixme : if values n, n+ are same, use repeat
        if start is None:
            start = 0
        else:
            start = self._translate_x_from_any(start)
        if start == 0 and (len(values) >= self.width):
            self.clear()
            cells = [
                Cell(value, style=style, cell_type=cell_type, currency=currency)
                for value in values
            ]
            self.extend_cells(cells)
        else:
            x = start
            for value in values:
                self.set_cell(
                    x,
                    Cell(value, style=style, cell_type=cell_type, currency=currency),
                    clone=False,
                )
                x += 1

    def rstrip(self, aggressive: bool = False) -> None:
        """Remove empty cells at the right of the row, in-place.

        An empty cell has no value but can have style. If `aggressive` is
        True, style is ignored.

        Args:
            aggressive: If True, ignores cell style.
        """
        for cell in reversed(self._get_cells()):
            if not cell.is_empty(aggressive=aggressive):
                break
            self.delete(cell)
        self._compute_row_cache()
        self._row_cache.clear_cell_indexes()

    def _current_length(self) -> int:
        """Return the current estimated length of the row.

        Returns:
            int: The length of the row.
        """
        idx_repeated_seq = self.elements_repeated_sequence(
            _XPATH_CELL, "table:number-columns-repeated"
        )
        repeated = [item[1] for item in idx_repeated_seq]
        if repeated:
            return sum(repeated)
        return 1

    def minimized_width(self) -> int:
        """Return the length of the row if the last repeated sequence is
        reduced to one.

        Returns:
            int: The minimized width of the row.
        """
        idx_repeated_seq = self.elements_repeated_sequence(
            _XPATH_CELL, "table:number-columns-repeated"
        )
        repeated = [item[1] for item in idx_repeated_seq]
        if repeated:
            cell = self.last_cell()
            if cell is not None and cell.is_empty(aggressive=True):
                repeated[-1] = 1
            min_width = sum(repeated)
        else:
            min_width = 1
        self._compute_row_cache()
        self._row_cache.clear_cell_indexes()
        return min_width

    def last_cell(self) -> Cell | None:
        """Return the last cell of the row.

        Returns:
            Cell or None: The last cell, or None if the row is empty.
        """
        try:
            return self._get_cells()[-1]
        except IndexError:
            return None

    def force_width(self, width: int) -> None:
        """Change the repeated property of the last cell of the row to comply
        with the required max width.

        Args:
            width: The target width.
        """
        cell = self.last_cell()
        if cell is None or not cell.is_empty(aggressive=True):
            return
        repeated = cell.repeated
        if repeated is None:
            return
        # empty repeated cell
        delta = self._current_length() - width
        if delta > 0:
            cell._set_repeated(repeated - delta)
            self._compute_row_cache()

    def is_empty(self, aggressive: bool = False) -> bool:
        """Return whether every cell in the row is empty.

        An empty cell has no value (or the value evaluates to False) and no
        style. If `aggressive` is True, empty cells with style are considered
        empty.

        Args:
            aggressive: If True, ignores cell style.

        Returns:
            bool: True if the row is empty, False otherwise.
        """
        return all(cell.is_empty(aggressive=aggressive) for cell in self._get_cells())


register_element_class(Row)



#!/usr/bin/env python
"""Many examples of how to change paragraph (and in-paragraph) styles, either
by changing the paragraph style itself or by using Span to select parts
of the paragraph. Includes several ways to create or import styles.
"""

import os
from collections.abc import Iterator
from itertools import cycle
from pathlib import Path

from odfdo import Document, Element, Header, Paragraph, Style

_DOC_SEQUENCE = 340
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "change_styles"
DATA = Path(__file__).parent / "data"
LOREM = (DATA / "lorem.txt").read_text(encoding="utf8")
STYLED_SOURCE = "lpod_styles.odt"
TARGET_BEFORE = "document_before.odt"
TARGET_AFTER = "document_after.odt"


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def iter_lorem() -> Iterator[str]:
    """Return infinite iterator on Lorem Ipsum content."""
    return cycle(lr.strip() for lr in LOREM.replace("\n", " ").split("."))


def make_base_document() -> Document:
    """Generate document from parts of lorem ipsum content."""
    document = Document("odt")
    body = document.body
    body.clear()
    # Add some content with headers
    lorem = iter_lorem()
    title1 = Header(1, next(lorem))
    body.append(title1)
    for _i in range(3):
        title = Header(2, next(lorem))
        body.append(title)
        for _j in range(5):
            body.append(Paragraph(next(lorem) + ". " + next(lorem) + "."))
    return document


def add_some_styles(document) -> None:
    """Add programmatically generated styles to the document."""
    # Always simpler to copy styles from an actual .odt existing file, but:
    document.insert_style(
        Style(
            family="paragraph",
            area="text",
            display_name="bold-blue",
            color="blue",
            bold=True,
        ),
        automatic=True,
    )
    document.insert_style(
        Style(
            family="paragraph",
            area="text",
            display_name="italic-red",
            color="red",
            bold=True,
            italic=True,
        ),
        automatic=True,
    )
    document.insert_style(
        Style(
            family="text",
            area="text",
            display_name="green",
            background_color="green",
        ),
        automatic=True,
    )
    document.insert_style(
        Style(
            family="text",
            area="text",
            display_name="bold-yellow-blue",
            color="yellow",
            background_color="blue",
            bold=True,
        ),
        automatic=True,
    )
    document.insert_style(
        Style(
            family="text",
            area="text",
            display_name="bold-white-black",
            color="white",
            background_color="black",
            bold=True,
        ),
        automatic=True,
    )
    document.insert_style(
        Style(
            family="text",
            area="text",
            display_name="italic-red-yellow",
            color="red",
            background_color="yellow",
            bold=True,
            italic=True,
        ),
        automatic=True,
    )


def add_style_from_xml(document: Document) -> None:
    """Add styles defined by XML content to the document."""
    # Styles can be defined by WML definition
    document.insert_style(
        Element.from_tag(
            '<style:style style:name="custom" '
            'style:display-name="custom" '
            'style:family="paragraph" '
            'style:parent-style-name="Text">'
            '<style:paragraph-properties fo:margin-left="2cm"/>'
            '<style:text-properties fo:color="#808080" loext:opacity="100%" '
            'fo:font-size="16pt" fo:font-style="normal" '
            'style:text-underline-style="solid" '
            'style:text-underline-width="auto" '
            'style:text-underline-color="font-color" '
            'fo:font-weight="bold"/>'
            "</style:style>"
        )
    )


def import_style_from_other_doc(document: Document) -> None:
    """Add styles imported from another document to the document."""
    styled_doc = Document(DATA / STYLED_SOURCE)
    highlight = styled_doc.get_style("text", display_name="Yellow Highlight")
    document.insert_style(highlight, automatic=True)


def apply_styles(document: Document) -> None:
    """Apply some style changes to the document."""

    def change_all_headers() -> None:
        style = document.get_style(family="text", display_name="green")
        # header styles should include some hints about he numeration level
        # So, here we just prefer to apply style with a span
        for header in document.body.headers:
            header.set_span(style.name, offset=0)

    def change_all_paragraphs() -> None:
        style = document.get_style(family="paragraph", display_name="bold-blue")
        for para in document.body.paragraphs:
            para.style = style.name

    def change_some_paragraph() -> None:
        style = document.get_style(family="paragraph", display_name="italic-red")
        document.body.get_paragraph(3).style = style.name
        document.body.get_paragraph(5).style = style.name
        document.body.get_paragraph(7).style = style.name

    def apply_span_regex() -> None:
        yellow = document.get_style(family="text", display_name="bold-yellow-blue")
        white = document.get_style(family="text", display_name="bold-white-black")
        for para in document.body.paragraphs:
            para.set_span(yellow.name, regex=r"tortor|ipsum")
            para.set_span(white.name, regex=r"A\w+")

    def apply_span_offset() -> None:
        red = document.get_style(family="text", display_name="italic-red-yellow")
        para = document.body.get_paragraph(2)
        para.set_span(red.name, offset=9, length=22)

    def apply_custom_style() -> None:
        para = document.body.get_paragraph(13)
        para.style = "custom"

    def apply_imported_style() -> None:
        para = document.body.get_paragraph(14)
        style = document.get_style(family="text", display_name="Yellow Highlight")
        # feature: to not highlight spaces, make as many Spans as required:
        for start, end in para.search_all(r"\w+"):
            length = end - start
            para.set_span(style.name, offset=start, length=length)

    change_all_headers()
    change_all_paragraphs()
    change_some_paragraph()
    apply_span_regex()
    apply_span_offset()
    apply_custom_style()
    apply_imported_style()


def main() -> None:
    document = make_base_document()
    save_new(document, TARGET_BEFORE)
    add_some_styles(document)
    add_style_from_xml(document)
    import_style_from_other_doc(document)
    apply_styles(document)
    test_unit(document)
    save_new(document, TARGET_AFTER)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert len(list(document.body.paragraphs)) == 15
    for display_name in (
        "bold-blue",
        "italic-red",
        "custom",
    ):
        style = document.get_style(family="paragraph", display_name=display_name)
        assert document.get_styled_elements(style.name)
    for display_name in (
        "green",
        "bold-yellow-blue",
        "bold-white-black",
        "Yellow Highlight",
    ):
        style = document.get_style(family="text", display_name=display_name)
        assert document.get_styled_elements(style.name)
    style = document.get_style(family="text", display_name="Yellow Highlight")
    assert len(document.get_styled_elements(style.name)) == 21


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
"""Provides the XmlPart base class for handling XML parts within ODF documents.

This module abstracts the interaction with XML content, allowing for easy
access, manipulation, and serialization of specific XML parts of an ODF file.
"""

from __future__ import annotations

from copy import deepcopy
from io import BytesIO
from typing import TYPE_CHECKING

from lxml.etree import _Element, _ElementTree, parse, tostring

from .container import Container, pretty_indent
from .element import Element, EText

if TYPE_CHECKING:
    from .body import Body
    from .settings import OfficeSettings


class XmlPart:
    """Represents an XML part within an ODF document.

    This class provides an abstraction layer over the underlying XML library
    (lxml), allowing for easier manipulation of XML content within ODF
    parts (e.g., 'content.xml', 'styles.xml').

    Attributes:
        part_name (str): The name of the XML part (e.g., "content.xml").
        container (Container): The ODF container associated with this XML part.
    """

    def __init__(self, part_name: str, container: Container) -> None:
        """Initializes an XmlPart instance.

        Args:
            part_name: The name of the XML part (e.g., "content.xml").
            container: The ODF container (zip file) that holds
                this XML part.
        """
        self.part_name = part_name
        self.container = container

        # Internal state
        self.__tree: _ElementTree | None = None
        self.__root: Element | None = None

    def _get_tree(self) -> _ElementTree:
        """Loads and returns the XML tree for the part.

        If the tree has not been loaded yet, it reads the part from the
        container and parses it.

        Returns:
            _ElementTree: The parsed XML ElementTree object.
        """
        if self.__tree is None:
            part = self.container.get_part(self.part_name)
            self.__tree = parse(BytesIO(part))  # type: ignore[arg-type]
        return self.__tree

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} part_name={self.part_name}>"

    # Public API

    @property
    def root(self) -> Element:
        """The root Element of this XML part.

        When accessed for the first time, it loads and parses the XML part
        into an Element object.
        """
        if self.__root is None:
            tree = self._get_tree()
            self.__root = Element.from_tag(tree.getroot())
        return self.__root

    def _get_body(self) -> Body:
        """Retrieves the document body ('office:body') from the root element.

        Returns:
            Body: The document body element.

        Raises:
            TypeError: If no 'office:body' element is found in the part.
        """
        body = self.root.document_body
        if not isinstance(body, Element):
            raise TypeError(f"No body found in {self.part_name!r}")
        return body

    @property
    def body(self) -> Body | OfficeSettings:
        """The document body element ('office:body').

        This property provides access to the main content body of the XML part.
        """
        return self._get_body()

    @body.setter
    def body(self, new_body: Element) -> None:
        """Sets the document body with a new Element.

        Args:
            new_body: The new 'office:body' element to set.
        """
        body = self._get_body()
        tail = body.tail
        body.clear()
        for item in new_body.children:
            body.append(item)
        if tail:  # pragma: nocover
            body.tail = tail

    def get_elements(self, xpath_query: str) -> list[Element]:
        """Returns a list of elements matching the XPath query.

        The XPath query is applied to the root of this XML part.

        Args:
            xpath_query: The XPath query string.

        Returns:
            list[Element]: A list of matching Element objects.
        """
        return self.root.get_elements(xpath_query)

    def get_element(self, xpath_query: str) -> Element | None:
        """Returns the first element matching the XPath query.

        The XPath query is applied to the root of this XML part.

        Args:
            xpath_query: The XPath query string.

        Returns:
            Element | None: The first matching Element object, or None if
                no match is found.
        """
        return self.root.get_element(xpath_query)

    def delete_element(self, child: Element) -> None:
        """Deletes a specified child element from the XML tree.

        Args:
            child: The child element to delete.
        """
        child.delete()

    def xpath(self, xpath_query: str) -> list[Element | EText]:
        """Applies an XPath query to the root of the XML part and its subtree.

        Args:
            xpath_query: The XPath query string.

        Returns:
            list[Element | EText]: A list of Element or EText instances
                matching the query.
        """
        return self.root.xpath(xpath_query)

    @property
    def clone(self) -> XmlPart:
        """Creates a deep copy of the XmlPart instance.

        The cloned part will have its own independent XML tree.

        Returns:
            XmlPart: A new XmlPart instance that is a clone of the original.
        """
        clone = object.__new__(self.__class__)
        for name in self.__dict__:
            if name == "container":
                setattr(clone, name, self.container.clone)
            elif name in ("_XmlPart__tree",):
                setattr(clone, name, None)
            else:
                value = getattr(self, name)
                value = deepcopy(value)
                setattr(clone, name, value)
        return clone

    def serialize(self, pretty: bool = False) -> bytes:
        """Serializes the XML part to bytes.

        Args:
            pretty: If True, the output XML will be pretty-printed.
                Defaults to False.

        Returns:
            bytes: The XML content as bytes, including the XML declaration.
        """
        if pretty:
            return self.pretty_serialize()
        xml_header = b'<?xml version="1.0" encoding="UTF-8"?>\n'
        tree = self._get_tree()
        bytes_tree = tostring(tree, encoding="unicode").encode("utf8")
        return xml_header + bytes_tree  # type: ignore[no-any-return]

    def pretty_serialize(self) -> bytes:
        """Serializes the XML part to bytes with pretty-printing.

        Returns:
            bytes: The pretty-printed XML content as bytes, including the
                XML declaration.
        """
        xml_header = b'<?xml version="1.0" encoding="UTF-8"?>\n'
        bytes_tree = tostring(
            self.custom_pretty_tree(),
            encoding="unicode",
        ).encode("utf8")
        return xml_header + bytes_tree  # type: ignore[no-any-return]

    def custom_pretty_tree(self) -> _ElementTree | _Element:
        """Returns a pretty-printed version of the XML tree.

        This method applies custom indentation for readability.

        Returns:
            _ElementTree | _Element: The pretty-printed XML tree or its root.
        """
        tree = self._get_tree()
        root = tree.getroot()
        return pretty_indent(root)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2012 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""RowGroup class for "table:table-row-group" tag."""

from __future__ import annotations

from typing import Any

from .element import Element, register_element_class
from .row import Row


class RowGroup(Element):
    """A group of rows  with common properties, "table:table-row-group".

    Partial implementation.

    The "table:table-row-group" element groups adjacent table rows. Every row
    group can contain header rows, and nested row groups. A row group can be
    visible or hidden.
    """

    # TODO
    _tag = "table:table-row-group"

    def __init__(
        self,
        height: int | None = None,
        width: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a RowGroup element (`table:table-row-group`).

        This element can be optionally pre-filled with a specified number
        of rows and cells.

        Args:
            height: The number of rows to create within the group.
            width: The number of cells to create in each new row.
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)
        if self._do_init and height is not None:
            for _i in range(height):
                row = Row(width=width)
                self.append(row)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"


register_element_class(RowGroup)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormCheckbox


@pytest.fixture
def form_checkbox() -> Iterable[FormCheckbox]:
    yield FormCheckbox(
        name="some name",
        title="some title",
        value="some value",
        control_implementation="some implem",
        current_state="checked",
        data_field="some data field",
        disabled=False,
        image_align="start",
        image_position="top",
        is_tristate=False,
        label="some label",
        linked_cell="A1",
        printable=True,
        state="checked",
        tab_index=4,
        tab_stop=False,
        visual_effect="flat",
        xml_id="control1",
    )


def test_form_checkbox_class():
    form = FormCheckbox()
    assert isinstance(form, FormCheckbox)


def test_form_checkbox_minimal_tag():
    form = Element.from_tag("<form:checkbox/>")
    assert isinstance(form, FormCheckbox)


def test_form_checkbox_serialize():
    form = FormCheckbox()
    assert form.serialize() == "<form:checkbox/>"


def test_form_checkbox_repr():
    form = FormCheckbox()
    assert repr(form) == "<FormCheckbox name=None xml_id=None>"


def test_form_checkbox_name(form_checkbox):
    assert form_checkbox.name == "some name"


def test_form_checkbox_title(form_checkbox):
    assert form_checkbox.title == "some title"


def test_form_checkbox_value(form_checkbox):
    assert form_checkbox.value == "some value"


def test_form_checkbox_control_implementation(form_checkbox):
    assert form_checkbox.control_implementation == "some implem"


def test_form_checkbox_current_state(form_checkbox):
    assert form_checkbox.current_state == "checked"


def test_form_checkbox_data_field(form_checkbox):
    assert form_checkbox.data_field == "some data field"


def test_form_checkbox_disabled(form_checkbox):
    assert form_checkbox.disabled is False


def test_form_checkbox_image_align(form_checkbox):
    assert form_checkbox.image_align == "start"


def test_form_checkbox_image_position(form_checkbox):
    assert form_checkbox.image_position == "top"


def test_form_checkbox_is_tristate(form_checkbox):
    assert form_checkbox.is_tristate is False


def test_form_checkbox_label(form_checkbox):
    assert form_checkbox.label == "some label"


def test_form_checkbox_linked_cell(form_checkbox):
    assert form_checkbox.linked_cell == "A1"


def test_form_checkbox_printable(form_checkbox):
    assert form_checkbox.printable is True


def test_form_checkbox_state(form_checkbox):
    assert form_checkbox.state == "checked"


def test_form_checkbox_tab_index(form_checkbox):
    assert form_checkbox.tab_index == "4"


def test_form_checkbox_tab_stop(form_checkbox):
    assert form_checkbox.tab_stop is False


def test_form_checkbox_visual_effect(form_checkbox):
    assert form_checkbox.visual_effect == "flat"


def test_form_checkbox_xml_id(form_checkbox):
    assert form_checkbox.xml_id == "control1"


def test_form_checkbox_form_id(form_checkbox):
    assert form_checkbox.form_id == "control1"


def test_form_checkbox_xforms_bind():
    form = FormCheckbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

from odfdo.utils import make_xpath_query


def test_element():
    query = make_xpath_query("descendant::text:p")
    assert query == "descendant::text:p"


def test_attribute():
    query = make_xpath_query("descendant::text:p", text_style="Standard")
    assert query == 'descendant::text:p[@text:style-name="Standard"]'


def test_two_attributes():
    query = make_xpath_query(
        "descendant::text:h", text_style="Standard", outline_level=1
    )
    expected = (
        'descendant::text:h[@text:outline-level="1"][@text:style-name="Standard"]'
    )
    assert query == expected


def test_position():
    query = make_xpath_query("descendant::text:h", position=1)
    assert query == "(descendant::text:h)[2]"


def test_attribute_position():
    query = make_xpath_query("descendant::text:p", text_style="Standard", position=1)
    assert query == '(descendant::text:p[@text:style-name="Standard"])[2]'


def test_two_attributes_position():
    query = make_xpath_query(
        "descendant::text:h", text_style="Standard", outline_level=1, position=1
    )
    expected = (
        '(descendant::text:h[@text:outline-level="1"][@text:style-name="Standard"])[2]'
    )
    assert query == expected


def test_case_presentation_class():
    query = make_xpath_query("descendant::text:p", presentation_class="foo")
    expected = 'descendant::text:p[@presentation:class="foo"]'
    assert query == expected


def test_case_true_arguent():
    query = make_xpath_query("descendant::text:p", change_id=True)
    expected = "descendant::text:p[@text:change-id]"
    assert query == expected


def test_case_last_arg():
    query = make_xpath_query("descendant::text:p", position=-1)
    expected = "(descendant::text:p)[last()]"
    assert query == expected


def test_case_last_2_arg():
    query = make_xpath_query("descendant::text:p", position=-2)
    expected = "(descendant::text:p)[last()-1]"
    assert query == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormColumn


@pytest.fixture
def form_column() -> Iterable[FormColumn]:
    yield FormColumn(
        name="some name",
        control_implementation="some implem",
        label="some label",
        text_style_name="style name",
    )


def test_form_column_class():
    form = FormColumn()
    assert isinstance(form, FormColumn)


def test_form_column_minimal_tag():
    form = Element.from_tag("<form:column/>")
    assert isinstance(form, FormColumn)


def test_form_column_serialize():
    form = FormColumn()
    assert form.serialize() == "<form:column/>"


def test_form_column_repr():
    form = FormColumn()
    assert repr(form) == "<FormColumn name=None>"


def test_form_column_name(form_column):
    assert form_column.name == "some name"


def test_form_column_control_implementation(form_column):
    assert form_column.control_implementation == "some implem"


def test_form_column_label(form_column):
    assert form_column.label == "some label"


def test_form_column_text_style_name(form_column):
    assert form_column.text_style_name == "style name"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
"""Container class, ODF file management."""

from __future__ import annotations

import base64
import contextlib
import io
import os
import shutil
import sys
import textwrap
import time
from copy import deepcopy
from functools import cache
from pathlib import Path, PurePath
from zipfile import ZIP_DEFLATED, ZIP_STORED, BadZipfile, ZipFile, is_zipfile

from lxml.etree import (
    Element,
    XMLSyntaxError,
    _Element,
    _ElementTree,
    fromstring,
    tostring,
)

from .const import (
    FOLDER,
    ODF_CONTENT,
    ODF_EXTENSIONS,
    ODF_FLAT_EXTENSIONS,
    ODF_MANIFEST,
    ODF_META,
    ODF_MIMETYPE_TO_FLAT_EXTENSION,
    ODF_MIMETYPES,
    ODF_SETTINGS,
    ODF_STYLES,
    OFFICE_PREFIX,
    OFFICE_VERSION,
    PACKAGING,
    XML,
    ZIP,
)
from .element import NAMESPACES_XML, xpath_compile
from .utils import bytes_to_str, str_to_bytes

TAB = "  "
TEXT_CONTENT = {
    "config:config-item",
    "db:data-source-setting-value",
    "db:table-filter-pattern",
    "db:table-type",
    "dc:creator",
    "dc:date",
    "dc:description",
    "dc:language",
    "dc:subject",
    "dc:title",
    "form:item",
    "form:option",
    "math:math",
    "meta:creation-date",
    "meta:date-string",
    "meta:editing-cycles",
    "meta:editing-duration",
    "meta:generator",
    "meta:initial-creator",
    "meta:keyword",
    "meta:print-date",
    "meta:printed-by",
    "meta:user-defined",
    "number:currency-symbol",
    "number:embedded-text",
    "number:text",
    "office:script",
    "presentation:date-time-decl",
    "presentation:footer-decl",
    "presentation:header-decl",
    "svg:desc",
    "svg:title",
    "table:desc",
    "table:title",
    "text:a",
    "text:author-initials",
    "text:author-name",
    "text:bibliography-mark",
    "text:bookmark-ref",
    "text:chapter",
    "text:character-count",
    "text:conditional-text",
    "text:creation-date",
    "text:creation-time",
    "text:creator",
    "text:database-display",
    "text:database-name",
    "text:database-row-number",
    "text:date",
    "text:dde-connection",
    "text:description",
    "text:editing-cycles",
    "text:editing-duration",
    "text:execute-macro",
    "text:expression",
    "text:file-name",
    "text:h",
    "text:hidden-paragraph",
    "text:hidden-text",
    "text:image-count",
    "text:index-entry-span",
    "text:index-title-template",
    "text:initial-creator",
    "text:keywords",
    "text:line-break",
    "text:linenumbering-separator",
    "text:measure",
    "text:meta",
    "text:meta-field",
    "text:modification-date",
    "text:modification-time",
    "text:note-citation",
    "text:note-continuation-notice-backward",
    "text:note-continuation-notice-forward",
    "text:note-ref",
    "text:number",
    "text:object-count",
    "text:p",
    "text:page-continuation",
    "text:page-count",
    "text:page-number",
    "text:page-variable-get",
    "text:page-variable-set",
    "text:paragraph-count",
    "text:placeholder",
    "text:print-date",
    "text:print-time",
    "text:printed-by",
    "text:reference-ref",
    "text:ruby-base",
    "text:ruby-text",
    "text:s",
    "text:script",
    "text:sender-city",
    "text:sender-company",
    "text:sender-country",
    "text:sender-email",
    "text:sender-fax",
    "text:sender-firstname",
    "text:sender-initials",
    "text:sender-lastname",
    "text:sender-phone-private",
    "text:sender-phone-work",
    "text:sender-position",
    "text:sender-postal-code",
    "text:sender-state-or-province",
    "text:sender-street",
    "text:sender-title",
    "text:sequence",
    "text:sequence-ref",
    "text:sheet-name",
    "text:soft-page-break",
    "text:span",
    "text:subject",
    "text:tab",
    "text:table-count",
    "text:table-formula",
    "text:template-name",
    "text:text-input",
    "text:time",
    "text:title",
    "text:user-defined",
    "text:user-field-get",
    "text:user-field-input",
    "text:variable-get",
    "text:variable-input",
    "text:variable-set",
    "text:word-count",
}

XML_TAG = b'<?xml version="1.0" encoding="UTF-8"?>\n'
NS_OFFICE = "urn:oasis:names:tc:opendocument:xmlns:office:1.0"


def printwarn(message: str) -> None:
    """Print a message on stderr.

    Args:
        message: the text to print.
    """
    print(f"Warning: {message}", file=sys.stderr)


@cache
def _ns_tag(tag: str) -> str:
    """Return the tag prefixed by NS_OFFICE."""
    return f"{{{NS_OFFICE}}}{tag}"


def pretty_indent(
    elem: _ElementTree | _Element,
    level: int = 0,
    ending_level: int = 0,
    textual_parent: bool = False,
) -> _ElementTree | _Element:
    """Return an indented _ElementTree.

    Args:
        elem: the source Element.
        level: current indentation level.
        ending_level: previous indentation level.
        textual_parent: True if the parent element contains text.

    """
    nb_child = len(elem)
    follow_level = level + 1
    tag = f"{elem.prefix}:{elem.tag.rpartition('}')[2]}"
    if tag in TEXT_CONTENT:
        is_textual = True
        if not textual_parent:
            elem.tail = "\n" + ending_level * TAB
    elif tag == "office:binary-data":
        is_textual = True
        elem.text = (
            textwrap.fill(
                elem.text,
                initial_indent=(ending_level + 11) * TAB,
                subsequent_indent=(follow_level * TAB)[:-1],
                width=79,
            ).lstrip()
            + "\n"
            + (follow_level - 1) * TAB
        )
        if not elem.tail:
            elem.tail = "\n" + ending_level * TAB
    else:
        is_textual = False
        if not textual_parent:
            elem.tail = "\n" + ending_level * TAB
            if nb_child > 0:
                elem.text = "\n" + follow_level * TAB
        else:
            if nb_child > 0:
                elem.text = (elem.text or "") + "\n" + follow_level * TAB
            if not elem.tail:
                elem.tail = "\n" + ending_level * TAB
    if nb_child > 0:
        for sub_elem in elem[:-1]:
            pretty_indent(sub_elem, follow_level, follow_level, is_textual)
        pretty_indent(elem[-1], follow_level, level, is_textual)
    return elem


def normalize_path(path: str) -> str:
    """Normalize a string path to Posix format.

    Args:
        path: The path to normalize.

    Returns:
        str: Posix representation of the path."""
    if path.endswith("/"):  # folder
        return PurePath(path[:-1]).as_posix() + "/"
    return PurePath(path).as_posix()


class Container:
    """Storage of the ODF document, as zip or other format."""

    def __init__(self, path: Path | str | io.BytesIO | None = None) -> None:
        """Storage of the ODF document, as zip or other format.

        Args:
            path: The path to the ODF file, a file-like object, or None.
        """
        self.__parts: dict[str, bytes | None] = {}
        self.__parts_ts: dict[str, int] = {}
        self.__path_like: Path | str | io.BytesIO | None = None
        self.__packaging: str = ZIP
        self.path: Path | None = None  # or Path
        if path:
            self.open(path)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.mimetype} path={self.path}>"

    def open(self, path_or_file: Path | str | io.BytesIO) -> None:
        """Load the content of an ODF file.

        Args:
            path_or_file: Path to the document, or an opened file.
        """
        self.__path_like = path_or_file
        if isinstance(path_or_file, (str, Path)):
            self.path = Path(path_or_file).expanduser()
            if not self.path.exists():
                raise FileNotFoundError(str(self.path))
            self.__path_like = self.path
        if (self.path or isinstance(self.__path_like, io.BytesIO)) and is_zipfile(
            self.__path_like
        ):
            self.__packaging = ZIP
            return self._read_zip()
        if self.path:
            is_folder = False
            with contextlib.suppress(OSError):
                is_folder = self.path.is_dir()
            if is_folder:
                self.__packaging = FOLDER
                return self._read_folder()
            # Check if it's a flat XML file
            with contextlib.suppress(OSError, ValueError):
                content = self.path.read_bytes()
                if self._is_flat_xml(content):
                    self.__packaging = XML
                    return self._read_xml(content)
        # Check if it's a BytesIO with flat XML
        if isinstance(self.__path_like, io.BytesIO):
            content = self.__path_like.getvalue()
            if self._is_flat_xml(content):
                self.__packaging = XML
                return self._read_xml(content)
        msg = f"Document format not managed by odfdo: {type(path_or_file)}."
        raise TypeError(msg)

    @staticmethod
    def _is_flat_xml(content: bytes) -> bool:
        """Check if content is a Flat ODF XML file.

        Args:
            content: The file content to check.

        Returns:
            True if the content appears to be a Flat ODF XML file.
        """
        # Must start with XML declaration or be parseable XML
        if not content.strip():
            return False
        if not content.lstrip().startswith(b"<?xml"):
            return False
        # Quick check for office:document element
        if b"office:document" not in content:
            return False
        try:
            root = fromstring(content)
        except (
            XMLSyntaxError,
            ValueError,
            TypeError,
        ):
            return False
        if root.tag != _ns_tag("document"):
            return False
        # Check for office:mimetype attribute
        mimetype = root.get(_ns_tag("mimetype"))
        if mimetype in ODF_MIMETYPES:
            return True
        # And accept if it has office:body child
        return root.find(_ns_tag("body")) is not None

    def _read_zip(self) -> None:
        if isinstance(self.__path_like, io.BytesIO):
            self.__path_like.seek(0)
        with ZipFile(self.__path_like) as zf:  # type: ignore
            mimetype = bytes_to_str(zf.read("mimetype"))
            if mimetype not in ODF_MIMETYPES:
                raise ValueError(f"Document of unknown type {mimetype}")
            self.__parts["mimetype"] = str_to_bytes(mimetype)
        if self.path is None:
            if isinstance(self.__path_like, io.BytesIO):
                self.__path_like.seek(0)
            # read the full file at once and forget file
            with ZipFile(self.__path_like) as zf:  # type: ignore
                for name in zf.namelist():
                    upath = normalize_path(name)
                    self.__parts[upath] = zf.read(name)
            self.__path_like = None

    def _read_folder(self) -> None:
        try:
            mimetype, timestamp = self._get_folder_part("mimetype")
        except OSError:
            msg = "Corrupted or not an OpenDocument folder (missing mimetype)"
            printwarn(msg)
            mimetype = b""
            timestamp = int(time.time())
        if bytes_to_str(mimetype) not in ODF_MIMETYPES:
            # Try to detect from content.xml if available
            detected = self._detect_mimetype_from_folder()
            if detected:
                msg = f"Document of unknown type {mimetype!r}, detected as {detected}."
                printwarn(msg)
                mimetype = str_to_bytes(detected)
            else:
                msg = f"Document of unknown type {mimetype!r}, try with ODF Text."
                printwarn(msg)
                mimetype = str_to_bytes(ODF_EXTENSIONS["odt"])
            self.__parts["mimetype"] = mimetype
            self.__parts_ts["mimetype"] = timestamp

    @staticmethod
    def _detect_image_mime_type(data: bytes) -> str:
        """Detect image mime type from file content."""
        if data.startswith(b"\xff\xd8\xff"):
            return "image/jpeg"
        if data.startswith(b"\x89PNG\r\n\x1a\n"):
            return "image/png"
        if data.startswith(b"GIF87a") or data.startswith(b"GIF89a"):
            return "image/gif"
        if data.startswith(b"<?xml") or b"<svg" in data[:100]:
            return "image/svg+xml"
        if data.startswith(b"BM"):
            return "image/bmp"
        if data.startswith(b"RIFF") and data[8:12] == b"WEBP":
            return "image/webp"
        return "application/octet-stream"

    @staticmethod
    def _image_mime_type_to_ext(mime_type: str) -> str:
        """Convert mime type to file extension."""
        mapping = {
            "image/jpeg": "jpg",
            "image/png": "png",
            "image/gif": "gif",
            "image/svg+xml": "svg",
            "image/bmp": "bmp",
            "image/webp": "webp",
            "image/tiff": "tiff",
        }
        return mapping.get(mime_type, "bin")

    @staticmethod
    def _suffix_to_mime_type(
        suffix: str,
        default: str = "application/octet-stream",
    ) -> str:
        """Convert file extension to mime type."""
        mapping = {
            ".xml": "text/xml",
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".png": "image/png",
            ".gif": "image/gif",
            ".svg": "image/svg+xml",
        }
        return mapping.get(suffix.lower(), default)

    def _read_xml(self, content: bytes) -> None:
        """Read a Flat ODF XML file and extract parts.

        Args:
            content: The XML content as bytes.
        """
        root = fromstring(content)
        mimetype, original_nsmap = self._extract_mimetype_and_namespaces(root)
        self.__parts["mimetype"] = str_to_bytes(mimetype)

        content_root, styles_root = self._create_document_roots(original_nsmap)
        master_style_refs = self._collect_master_style_refs(root)
        image_parts = self._distribute_elements(
            root, content_root, styles_root, master_style_refs
        )

        # Process embedded content
        xlink_ns = "{http://www.w3.org/1999/xlink}"
        image_counter = self._process_embedded_images(
            content_root, styles_root, image_parts, xlink_ns
        )
        image_counter = self._process_fill_images(
            content_root, styles_root, image_parts, image_counter, xlink_ns
        )
        self._process_embedded_objects(content_root, xlink_ns)
        self._process_form_images(content_root, image_parts)

        # Store image parts
        for path, data in image_parts.items():
            if path not in self.__parts:
                self.__parts[path] = data

        self._serialize_documents(content_root, styles_root)
        self._create_manifest()

    def _detect_mimetype_from_content(self, root: _Element) -> str:
        """Detect mimetype from body content when attribute is missing.

        Args:
            root: The root element of the parsed XML.

        Returns:
            The detected mimetype string.
        """
        body = root.find(_ns_tag("body"))
        if body is not None:
            if body.find(_ns_tag("spreadsheet")) is not None:
                return ODF_EXTENSIONS["ods"]
            if body.find(_ns_tag("presentation")) is not None:
                return ODF_EXTENSIONS["odp"]
            if body.find(_ns_tag("drawing")) is not None:
                return ODF_EXTENSIONS["odg"]
            if body.find(_ns_tag("chart")) is not None:
                return ODF_EXTENSIONS["odc"]
            if body.find(_ns_tag("image")) is not None:
                return ODF_EXTENSIONS["odi"]
            if body.find(_ns_tag("formula")) is not None:
                return ODF_EXTENSIONS["odf"]
        return ODF_EXTENSIONS["odt"]

    def _detect_mimetype_from_folder(self) -> str | None:
        """Detect mimetype by examining content.xml in folder.

        Returns:
            The detected mimetype string or None if detection fails.
        """
        try:
            if self.path is None:
                return None
            content_path = self.path / "content.xml"
            if not content_path.exists():
                return None
            root = fromstring(content_path.read_bytes())
            return self._detect_mimetype_from_content(root)
        except Exception:
            return None

    def _extract_mimetype_and_namespaces(
        self, root: _Element
    ) -> tuple[str, dict[str, str]]:
        """Extract mimetype and namespaces from the root element.

        Args:
            root: The root element of the parsed XML.

        Returns:
            Tuple of (mimetype, original_nsmap).
        """
        mimetype = root.get(_ns_tag("mimetype"))
        if mimetype is None:
            mimetype = self._detect_mimetype_from_content(root)
        # Filter out None keys (default namespace) and manifest namespace
        original_nsmap = {
            prefix: uri
            for prefix, uri in root.nsmap.items()
            if prefix is not None
            and uri != "urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
        }
        return mimetype, original_nsmap

    def _create_document_roots(
        self, original_nsmap: dict[str, str]
    ) -> tuple[Element, Element]:
        """Create content and styles document roots.

        Args:
            original_nsmap: The original namespace map.

        Returns:
            Tuple of (content_root, styles_root).
        """
        content_root = Element(_ns_tag("document-content"), nsmap=original_nsmap)
        content_root.set(_ns_tag("version"), OFFICE_VERSION)

        styles_root = Element(_ns_tag("document-styles"), nsmap=original_nsmap)
        styles_root.set(_ns_tag("version"), OFFICE_VERSION)

        return content_root, styles_root

    def _collect_master_style_refs(self, root: _Element) -> set[str]:
        """Collect style names referenced from master-styles.

        These styles (like header/footer paragraph styles) need to stay in styles.xml.

        Args:
            root: The root element of the parsed XML.

        Returns:
            Set of style names referenced from master-styles.
        """
        ns_style = "{urn:oasis:names:tc:opendocument:xmlns:style:1.0}"
        ns_text = "{urn:oasis:names:tc:opendocument:xmlns:text:1.0}"
        ns_draw = "{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}"
        ns_presentation = "{urn:oasis:names:tc:opendocument:xmlns:presentation:1.0}"
        master_style_refs: set[str] = set()

        for child in root:
            if child.tag == _ns_tag("master-styles"):
                for elem in child.iter():
                    style_ref = elem.get(f"{ns_text}style-name")
                    if style_ref:
                        master_style_refs.add(style_ref)
                    style_name = elem.get(f"{ns_style}name")
                    if style_name:
                        master_style_refs.add(style_name)
                    draw_style = elem.get(f"{ns_draw}style-name")
                    if draw_style:
                        master_style_refs.add(draw_style)
                    pres_style = elem.get(f"{ns_presentation}style-name")
                    if pres_style:
                        master_style_refs.add(pres_style)

        return master_style_refs

    def _distribute_elements(
        self,
        root: _Element,
        content_root: Element,
        styles_root: Element,
        master_style_refs: set[str],
    ) -> dict[str, bytes]:
        """Distribute child elements to content or styles roots.

        Args:
            root: The root element of the parsed XML.
            content_root: The content document root.
            styles_root: The styles document root.
            master_style_refs: Set of style names referenced from master-styles.

        Returns:
            Dictionary of image parts.
        """
        ns_style = "{urn:oasis:names:tc:opendocument:xmlns:style:1.0}"
        image_parts: dict[str, bytes] = {}

        for child in root:
            tag = child.tag
            if tag == _ns_tag("meta"):
                self._process_meta_element(child)
            elif tag == _ns_tag("document-meta"):
                self._process_document_meta_element(child)
            elif tag == _ns_tag("settings"):
                self._process_settings_element(child)
            elif tag == _ns_tag("automatic-styles"):
                self._process_automatic_styles(
                    child, content_root, styles_root, master_style_refs, ns_style
                )
            elif tag in {
                _ns_tag("styles"),
                _ns_tag("master-styles"),
                _ns_tag("font-face-decls"),
            }:
                self._merge_or_append_to_styles(child, styles_root)
            else:
                content_root.append(child)

        return image_parts

    def _process_meta_element(self, meta: _Element) -> None:
        """Process meta element and store as meta.xml."""
        meta_root = Element(
            _ns_tag("document-meta"),
            nsmap={
                "office": NS_OFFICE,
                "meta": "urn:oasis:names:tc:opendocument:xmlns:meta:1.0",
                "dc": "http://purl.org/dc/elements/1.1/",
            },
        )
        meta_root.set(_ns_tag("version"), OFFICE_VERSION)
        meta_root.append(meta)
        meta_xml = XML_TAG + tostring(
            meta_root, encoding="UTF-8", xml_declaration=False
        )
        self.__parts[ODF_META] = meta_xml

    def _process_document_meta_element(self, doc_meta: _Element) -> None:
        """Process document-meta element (already wrapped) and store as meta.xml."""
        meta_xml = XML_TAG + tostring(doc_meta, encoding="UTF-8", xml_declaration=False)
        self.__parts[ODF_META] = meta_xml

    def _process_settings_element(self, settings: _Element) -> None:
        """Process settings element and store as settings.xml."""
        settings_xml = XML_TAG + (
            b"<office:document-settings xmlns:office="
            b'"urn:oasis:names:tc:opendocument:xmlns:office:1.0" '
            b'xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0">'
        )
        for settings_child in settings:
            settings_xml += tostring(
                settings_child, encoding="UTF-8", xml_declaration=False
            )
        settings_xml += b"</office:document-settings>"
        self.__parts[ODF_SETTINGS] = settings_xml

    def _process_automatic_styles(
        self,
        auto_styles: _Element,
        content_root: Element,
        styles_root: Element,
        master_style_refs: set[str],
        ns_style: str,
    ) -> None:
        """Distribute automatic-styles to content or styles roots.

        - Styles referenced from master-styles go to styles.xml
        - Page layouts go to styles.xml
        - Everything else goes to content.xml
        """
        content_auto = content_root.find(_ns_tag("automatic-styles"))
        if content_auto is None:
            content_auto = Element(_ns_tag("automatic-styles"))
            content_root.insert(0, content_auto)

        styles_auto = styles_root.find(_ns_tag("automatic-styles"))
        if styles_auto is None:
            styles_auto = Element(_ns_tag("automatic-styles"))
            styles_root.append(styles_auto)

        for grandchild in auto_styles:
            tag_name = grandchild.tag.split("}")[-1]
            style_name = grandchild.get(f"{ns_style}name")
            is_page_layout = tag_name == "page-layout"
            is_master_style = style_name is not None and style_name in master_style_refs

            if is_page_layout or is_master_style:
                styles_auto.append(grandchild)
            else:
                content_auto.append(grandchild)

    def _merge_or_append_to_styles(self, child: _Element, styles_root: Element) -> None:
        """Merge child into styles root or append if not exists."""
        tag = child.tag
        existing = styles_root.find(tag)
        if existing is not None:
            for grandchild in child:
                existing.append(grandchild)
        else:
            styles_root.append(child)

    def _process_embedded_images(
        self,
        content_root: Element,
        styles_root: Element,
        image_parts: dict[str, bytes],
        xlink_ns: str,
    ) -> int:
        """Process draw:image elements with office:binary-data.

        Returns:
            The updated image counter.
        """
        ns_draw = "{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}"
        xpath_expr = xpath_compile("descendant::draw:image[office:binary-data]")
        image_counter = 0

        for root_elem in (content_root, styles_root):
            for img_elem in xpath_expr(root_elem):
                binary_data_elem = img_elem.find(_ns_tag("binary-data"))
                if binary_data_elem is not None and binary_data_elem.text:
                    try:
                        image_data = base64.standard_b64decode(
                            binary_data_elem.text.strip()
                        )
                        mime_type = img_elem.get(f"{ns_draw}mime-type")
                        if mime_type is None or mime_type == "None":
                            mime_type = self._detect_image_mime_type(image_data)
                        ext = self._image_mime_type_to_ext(mime_type)

                        image_counter += 1
                        image_path = f"Pictures/image{image_counter}.{ext}"
                        image_parts[image_path] = image_data

                        img_elem.remove(binary_data_elem)
                        img_elem.set(f"{xlink_ns}href", image_path)
                        img_elem.set(f"{xlink_ns}type", "simple")
                        img_elem.set(f"{xlink_ns}show", "embed")
                        img_elem.set(f"{xlink_ns}actuate", "onLoad")
                    except Exception as e:
                        printwarn(f"Failed to decode embedded image: {e}")

        return image_counter

    def _process_fill_images(
        self,
        content_root: Element,
        styles_root: Element,
        image_parts: dict[str, bytes],
        image_counter: int,
        xlink_ns: str,
    ) -> int:
        """Process draw:fill-image elements with office:binary-data.

        Returns:
            The updated image counter.
        """
        xpath_fill_expr = xpath_compile(
            "descendant::draw:fill-image[office:binary-data]"
        )

        for root_elem in (content_root, styles_root):
            for fill_img_elem in xpath_fill_expr(root_elem):
                binary_data_elem = fill_img_elem.find(_ns_tag("binary-data"))
                if binary_data_elem is None or not binary_data_elem.text:
                    continue
                try:
                    image_data = base64.standard_b64decode(
                        binary_data_elem.text.strip()
                    )
                    mime_type = self._detect_image_mime_type(image_data)
                    ext = self._image_mime_type_to_ext(mime_type)

                    image_counter += 1
                    image_path = f"Pictures/image{image_counter}.{ext}"
                    image_parts[image_path] = image_data

                    fill_img_elem.remove(binary_data_elem)
                    fill_img_elem.set(f"{xlink_ns}href", image_path)
                    fill_img_elem.set(f"{xlink_ns}type", "simple")
                except Exception as e:
                    printwarn(f"Failed to decode embedded fill image: {e}")

        return image_counter

    def _process_embedded_objects(self, content_root: Element, xlink_ns: str) -> None:
        """Process draw:object elements with embedded content."""
        object_counter = 0
        xpath_obj_expr = xpath_compile("descendant::draw:object[office:document]")

        for obj_elem in xpath_obj_expr(content_root):
            try:
                object_counter += 1
                obj_path = f"Object {object_counter}"

                doc_wrapper = obj_elem.find(_ns_tag("document"))
                # Can not happen: XPath already filters for objects with document
                if doc_wrapper is None:  # pragma: no cover
                    continue

                # Create object content.xml from body
                body_elem = doc_wrapper.find(_ns_tag("body"))
                if body_elem is not None:
                    content_root_elem = Element(
                        _ns_tag("document-content"),
                        nsmap={"office": NS_OFFICE},
                    )
                    content_root_elem.set(_ns_tag("version"), OFFICE_VERSION)
                    for child in list(body_elem):
                        content_root_elem.append(child)

                    obj_content_xml = XML_TAG + tostring(
                        content_root_elem, encoding="UTF-8", xml_declaration=False
                    )
                    self.__parts[f"{obj_path}/content.xml"] = obj_content_xml

                # Extract styles
                auto_styles = doc_wrapper.find(_ns_tag("automatic-styles"))
                if auto_styles is not None:
                    styles_root_elem = Element(
                        _ns_tag("document-styles"),
                        nsmap={
                            "office": NS_OFFICE,
                            "style": "urn:oasis:names:tc:opendocument:xmlns:style:1.0",
                        },
                    )
                    styles_root_elem.set(_ns_tag("version"), OFFICE_VERSION)
                    styles_root_elem.append(auto_styles)

                    obj_styles_xml = XML_TAG + tostring(
                        styles_root_elem, encoding="UTF-8", xml_declaration=False
                    )
                    self.__parts[f"{obj_path}/styles.xml"] = obj_styles_xml

                # Extract meta
                meta_elem = doc_wrapper.find(_ns_tag("meta"))
                if meta_elem is not None:
                    meta_root_elem = Element(
                        _ns_tag("document-meta"),
                        nsmap={
                            "office": NS_OFFICE,
                            "meta": "urn:oasis:names:tc:opendocument:xmlns:meta:1.0",
                        },
                    )
                    meta_root_elem.set(_ns_tag("version"), OFFICE_VERSION)
                    for child in list(meta_elem):
                        meta_root_elem.append(child)

                    obj_meta_xml = XML_TAG + tostring(
                        meta_root_elem, encoding="UTF-8", xml_declaration=False
                    )
                    self.__parts[f"{obj_path}/meta.xml"] = obj_meta_xml

                # Replace the embedded object with a reference
                for child in list(obj_elem):
                    obj_elem.remove(child)
                obj_elem.set(f"{xlink_ns}href", f"./{obj_path}")
                obj_elem.set(f"{xlink_ns}type", "simple")
                obj_elem.set(f"{xlink_ns}show", "embed")
                obj_elem.set(f"{xlink_ns}actuate", "onLoad")

            except Exception as e:
                printwarn(f"Failed to extract embedded object: {e}")

    def _process_form_images(
        self, content_root: Element, image_parts: dict[str, bytes]
    ) -> None:
        """Process form elements with embedded image data."""
        ns_form = "{urn:oasis:names:tc:opendocument:xmlns:form:1.0}"
        xpath_form_img = xpath_compile("descendant::form:*[office:binary-data]")
        image_counter = 0

        for form_elem in xpath_form_img(content_root):
            binary_data_elem = form_elem.find(_ns_tag("binary-data"))
            if binary_data_elem is None or not binary_data_elem.text:
                continue
            try:
                image_data = base64.standard_b64decode(binary_data_elem.text.strip())
                mime_type = self._detect_image_mime_type(image_data)
                ext = self._image_mime_type_to_ext(mime_type)

                image_counter += 1
                image_path = f"Pictures/image{image_counter}.{ext}"
                image_parts[image_path] = image_data

                form_elem.remove(binary_data_elem)
                form_elem.set(f"{ns_form}image-data", image_path)
            except Exception as e:
                printwarn(f"Failed to decode embedded form image: {e}")

    def _serialize_documents(self, content_root: Element, styles_root: Element) -> None:
        """Serialize content.xml and styles.xml."""
        if len(content_root) > 0:
            content_xml = XML_TAG + tostring(
                content_root, encoding="UTF-8", xml_declaration=False
            )
            self.__parts[ODF_CONTENT] = content_xml

        if len(styles_root) > 0:
            styles_xml = XML_TAG + tostring(
                styles_root, encoding="UTF-8", xml_declaration=False
            )
            self.__parts[ODF_STYLES] = styles_xml

    def _create_manifest(self) -> None:
        """Create the META-INF/manifest.xml from current parts."""
        manifest_parts = [
            XML_TAG,
            b'<manifest:manifest xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0">\n',
        ]

        # Add mimetype entry
        mimetype = (
            self.__parts.get("mimetype") or b"application/vnd.oasis.opendocument.text"
        )

        manifest_parts.append(
            f'<manifest:file-entry manifest:media-type="{mimetype.decode()}" manifest:full-path="/"/>\n'.encode()
        )

        # Add entries for each part
        for path, data in self.__parts.items():
            if data is None or path == "mimetype":
                continue
            # Determine media type based on path
            media_type = self._suffix_to_mime_type(Path(path).suffix)
            manifest_parts.append(
                f'<manifest:file-entry manifest:media-type="{media_type}" manifest:full-path="{path}"/>\n'.encode()
            )

        manifest_parts.append(b"</manifest:manifest>")
        self.__parts[ODF_MANIFEST] = b"".join(manifest_parts)

    def _parse_folder(self, folder: str) -> list[str]:
        parts = []
        if self.path is None:
            raise ValueError("Document path is not defined")
        root = self.path / folder
        for path in root.iterdir():
            if path.name.startswith("."):  # no hidden files
                continue
            relative_path = path.relative_to(self.path)
            if path.is_file():
                parts.append(relative_path.as_posix())
            if path.is_dir():
                sub_parts = self._parse_folder(str(relative_path))
                if sub_parts:
                    parts.extend(sub_parts)
                else:
                    # store leaf directories
                    parts.append(relative_path.as_posix() + "/")
        return parts

    def _get_folder_parts(self) -> list[str]:
        """Get the list of members in the ODF folder."""
        return self._parse_folder("")

    def _get_folder_part(self, name: str) -> tuple[bytes, int]:
        """Get bytes of a part from the ODF folder, with timestamp."""
        if self.path is None:
            raise ValueError("Document path is not defined")
        path = self.path / name
        timestamp = int(path.stat().st_mtime)
        if path.is_dir():
            return (b"", timestamp)
        return (path.read_bytes(), timestamp)

    def _get_folder_part_timestamp(self, name: str) -> int:
        if self.path is None:
            raise ValueError("Document path is not defined")
        path = self.path / name
        try:
            timestamp = int(path.stat().st_mtime)
        except OSError:
            timestamp = -1
        return timestamp

    def _get_zip_part(self, name: str) -> bytes | None:
        """Get bytes of a part from the Zip ODF file.

        No cache.
        """
        if self.path is None:
            raise ValueError("Document path is not defined")
        try:
            with ZipFile(self.path) as zf:
                upath = normalize_path(name)
                self.__parts[upath] = zf.read(name)
                return self.__parts[upath]
        except BadZipfile:
            return None

    def _get_all_zip_part(self) -> None:
        """Read all parts.

        No cache.
        """
        if self.path is None:
            raise ValueError("Document path is not defined")
        try:
            with ZipFile(self.path) as zf:
                for name in zf.namelist():
                    upath = normalize_path(name)
                    self.__parts[upath] = zf.read(name)
        except BadZipfile:
            pass

    def _save_zip(self, target: str | Path | io.BytesIO) -> None:
        """Save a Zip ODF from the available parts."""
        parts = self.__parts
        with ZipFile(target, "w", compression=ZIP_DEFLATED) as filezip:
            # Parts to save, except manifest at the end
            part_names = list(parts.keys())
            try:
                part_names.remove(ODF_MANIFEST)
            except ValueError:
                printwarn(f"Missing '{ODF_MANIFEST}'")
            # "Pretty-save" parts in some order
            # mimetype requires to be first and uncompressed
            mimetype = parts.get("mimetype")
            if mimetype is None:
                raise ValueError("Mimetype is not defined")
            try:
                filezip.writestr("mimetype", mimetype, ZIP_STORED)
                part_names.remove("mimetype")
            except (
                ValueError,
                KeyError,
            ):
                printwarn("Missing 'mimetype'")
            # XML parts
            for path in ODF_CONTENT, ODF_META, ODF_SETTINGS, ODF_STYLES:
                if path not in parts:
                    printwarn(f"Missing '{path}'")
                    continue
                part = parts[path]
                if part is None:
                    continue
                filezip.writestr(path, part)
                part_names.remove(path)
            # Everything else
            for path in part_names:
                data = parts[path]
                if data is None:
                    # Deleted
                    continue
                filezip.writestr(path, data)
            with contextlib.suppress(KeyError):
                part = parts[ODF_MANIFEST]
                if part is not None:
                    filezip.writestr(ODF_MANIFEST, part)

    def _save_folder(self, folder: Path | str) -> None:
        """Save a folder ODF from the available parts."""

        def dump(part_path: str, content: bytes) -> None:
            if part_path.endswith("/"):  # folder
                is_folder = True
                pure_path = PurePath(folder, part_path[:-1])
            else:
                is_folder = False
                pure_path = PurePath(folder, part_path)
            path = Path(pure_path)
            if is_folder:
                path.mkdir(parents=True, exist_ok=True)
            else:
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_bytes(content)
                path.chmod(0o666)

        for part_path, data in self.__parts.items():
            if data is None:
                # Deleted
                continue
            dump(part_path, data)

    def _encoded_image(self, elem: _Element) -> _Element | None:
        mime_type = elem.get(
            "{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}mime-type"
        )
        path = elem.get("{http://www.w3.org/1999/xlink}href")
        if not path:
            return None
        path = path.lstrip("./")
        content = self.__parts[path]
        if not content:
            return None
        text = base64.standard_b64encode(content).decode()
        # Only include mime-type attribute if it's a valid value
        if mime_type and mime_type != "None":
            ebytes = (
                f'<draw:image draw:mime-type="{mime_type}">'
                f"<office:binary-data>{text}\n</office:binary-data></draw:image>"
            ).encode()
        else:
            ebytes = (
                f"<draw:image>"
                f"<office:binary-data>{text}\n</office:binary-data></draw:image>"
            ).encode()
        root = fromstring(NAMESPACES_XML % ebytes)
        return root[0]

    def _embed_form_image_data(self, elem: _Element, image_path: str) -> None:
        """Embed image data for form elements with form:image-data attribute.

        Form elements (like image buttons) reference images via form:image-data
        attribute. We need to embed the image as base64 and replace the
        attribute reference with the embedded data.
        """
        # Strip leading "./" from path
        path = image_path.lstrip("./")
        content = self.__parts.get(path)
        if not content:
            return

        # Encode the image as base64
        text = base64.standard_b64encode(content).decode()

        # Create the office:binary-data element
        binary_data_elem = Element(_ns_tag("binary-data"))
        binary_data_elem.text = text + "\n"

        # Append the binary-data element to the form element
        elem.append(binary_data_elem)

    def _encoded_fill_image(self, elem: _Element) -> _Element | None:
        """Encode a draw:fill-image element with binary data.

        draw:fill-image elements reference images via xlink:href and are used
        for background images in presentations.
        """
        path = elem.get("{http://www.w3.org/1999/xlink}href")
        if not path:
            return None
        # Strip leading "./" from path (e.g., "./Pictures/image.jpg")
        path = path.lstrip("./")
        content = self.__parts[path]
        if not content:
            return None
        # mime_type = self._suffix_to_mime_type(Path(path).suffix, "image/jpeg")
        name = elem.get("{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}name")
        display_name = elem.get(
            "{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}display-name"
        )
        text = base64.standard_b64encode(content).decode()
        # Build the element with preserved attributes
        attrs = []
        if name:
            attrs.append(f'draw:name="{name}"')
        if display_name:
            attrs.append(f'draw:display-name="{display_name}"')
        attr_str = " ".join(attrs)
        if attr_str:
            attr_str = " " + attr_str
        ebytes = (
            f"<draw:fill-image{attr_str}>"
            f"<office:binary-data>{text}\n</office:binary-data></draw:fill-image>"
        ).encode()
        root = fromstring(NAMESPACES_XML % ebytes)
        return root[0]

    def _encoded_object(self, elem: _Element) -> _Element | None:
        """Encode a draw:object element by embedding the object content.

        draw:object elements reference embedded ODF objects (like charts)
        via xlink:href. In flat format, we need to embed the object content
        directly as an office:document structure.
        """
        path = elem.get("{http://www.w3.org/1999/xlink}href")
        if not path:
            return None
        # Strip leading "./" from path (e.g., "./Object 1")
        path = path.lstrip("./")

        # Check if we have parts for this object (Object 1/content.xml, etc.)
        content_path = f"{path}/content.xml"
        styles_path = f"{path}/styles.xml"
        meta_path = f"{path}/meta.xml"

        # If we don't have the object parts, return None to keep original reference
        if content_path not in self.__parts:
            return None

        # Parse the content.xml
        try:
            content_doc = fromstring(self.__parts[content_path])
        except Exception:
            return None

        # Create a new office:document element with the same namespace map
        # This preserves all namespace declarations from the original
        obj_doc = Element(_ns_tag("document"), nsmap=content_doc.nsmap)

        # Add required office:document attributes
        obj_doc.set(_ns_tag("mimetype"), "application/vnd.oasis.opendocument.chart")
        obj_doc.set(_ns_tag("version"), OFFICE_VERSION)

        # Force all namespace declarations by adding dummy attributes
        # lxml optimizes away namespace declarations if they're already in scope
        # We add them with a unique suffix to prevent optimization
        for prefix, uri in content_doc.nsmap.items():
            if prefix and prefix != "office":
                # Use a special attribute that won't conflict with real ones
                dummy_attr = f"{{{uri}}}__{prefix}__decl"
                obj_doc.set(dummy_attr, "")

        # Add meta if available
        if self.__parts.get(meta_path):
            try:
                meta_root = fromstring(self.__parts[meta_path])
                for child in meta_root:
                    obj_doc.append(child)
            except Exception as e:
                printwarn(f"Error in meta {e}")

        # Add empty styles (objects often don't have separate styles)
        styles_elem = Element(_ns_tag("styles"))
        obj_doc.append(styles_elem)

        # Add automatic-styles from styles.xml if available
        if self.__parts.get(styles_path):
            try:
                styles_root = fromstring(self.__parts[styles_path])
                auto_styles = styles_root.find(_ns_tag("automatic-styles"))
                if auto_styles is not None:
                    obj_doc.append(auto_styles)
            except Exception as e:
                printwarn(f"Error in styles path {e}")

        # Extract automatic-styles from content and add to obj_doc
        content_auto_styles = content_doc.find(_ns_tag("automatic-styles"))
        if content_auto_styles is not None:
            obj_doc.append(content_auto_styles)

        # Add body with content from content.xml
        body_elem = Element(_ns_tag("body"))
        content_body = content_doc.find(_ns_tag("body"))
        if content_body is not None:
            for child in content_body:
                body_elem.append(child)
        else:
            # If no office:body, add all children except automatic-styles
            for child in content_doc:
                if child.tag != _ns_tag("automatic-styles"):
                    body_elem.append(child)
        obj_doc.append(body_elem)

        # Create the draw:object element with the embedded content
        # To preserve namespace declarations, we need to serialize and re-parse
        # because lxml optimizes them away when appending to parent
        new_obj = Element("{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}object")

        # Serialize the embedded document to a string
        obj_doc_str = tostring(obj_doc, encoding="unicode", pretty_print=False)

        # Parse it back and append to new_obj
        # This preserves the namespace declarations from the original serialization
        obj_doc_parsed = fromstring(obj_doc_str.encode("utf-8"))
        new_obj.append(obj_doc_parsed)

        return new_obj

    def _xml_content(self, pretty: bool = True) -> bytes:
        mimetype_b = self.__parts["mimetype"]
        if mimetype_b is None:
            # use some default
            mimetype = "application/vnd.oasis.opendocument.text"
        else:
            mimetype = mimetype_b.decode("utf8")
        doc_xml = (
            OFFICE_PREFIX.decode("utf8")
            + f'office:version="{OFFICE_VERSION}"\n'
            + f'office:mimetype="{mimetype}">'
            + "</office:document>"
        )
        root = fromstring(doc_xml.encode("utf8"))
        for path in ODF_META, ODF_SETTINGS, ODF_STYLES, ODF_CONTENT:
            if path not in self.__parts:
                printwarn(f"Missing '{path}'")
                continue
            part = self.__parts[path]
            if part is None:
                continue
            if isinstance(part, bytes):
                xpart = fromstring(part)
            else:
                xpart = part
            if path in {ODF_CONTENT, ODF_STYLES}:
                # Process draw:image elements (both content and styles)
                xpath = xpath_compile("descendant::draw:image")
                images = xpath(xpart)
                if images:
                    for elem in images:
                        encoded = self._encoded_image(elem)
                        if encoded is not None:
                            elem.getparent().replace(elem, encoded)
                # Process draw:fill-image elements (used for background images in presentations)
                xpath_fill = xpath_compile("descendant::draw:fill-image")
                fill_images = xpath_fill(xpart)
                if fill_images:
                    for elem in fill_images:
                        encoded = self._encoded_fill_image(elem)
                        if encoded is not None:
                            elem.getparent().replace(elem, encoded)
                # Process draw:object elements (embedded objects like charts)
                if path == ODF_CONTENT:
                    xpath_obj = xpath_compile("descendant::draw:object")
                    objects = xpath_obj(xpart)
                    if objects:
                        for elem in objects:
                            encoded = self._encoded_object(elem)
                            if encoded is not None:
                                elem.getparent().replace(elem, encoded)
                    # Process form elements with form:image-data attribute
                    # (used for form control images like buttons with images)
                    ns_form = "{urn:oasis:names:tc:opendocument:xmlns:form:1.0}"
                    xpath_form = xpath_compile("descendant::form:*[@form:image-data]")
                    form_elems = xpath_form(xpart)
                    if form_elems:
                        for elem in form_elems:
                            image_path = elem.get(f"{ns_form}image-data")
                            if image_path:
                                self._embed_form_image_data(elem, image_path)
            for child in xpart:
                root.append(child)
        if pretty:
            bytes_tree: bytes = tostring(
                pretty_indent(root),
                encoding="unicode",
            ).encode("utf8")
            return XML_TAG + bytes_tree
        else:
            return tostring(  # type: ignore[no-any-return]
                root,
                encoding="UTF-8",
                xml_declaration=True,
            )

    def _save_xml(
        self,
        target: Path | str | io.BytesIO,
        pretty: bool = True,
    ) -> None:
        """Save a XML flat ODF format from the available parts."""
        if isinstance(target, (Path, str)):
            target_path = Path(target)
            # Preserve Flat ODF extension is exists, default to .xml
            if target_path.suffix.lower() not in ODF_FLAT_EXTENSIONS:
                target_path = target_path.with_suffix(".xml")
            target_path.write_bytes(self._xml_content(pretty))
        else:
            target.write(self._xml_content(pretty))

    # Public API

    def get_parts(self) -> list[str]:
        """Get the list of the parts in the Container.

        Returns:
            The list of path of the parts in the Container.
        """
        if not self.path:
            # maybe a file like zip archive or xml
            return list(self.__parts.keys())
        if self.__packaging == ZIP:
            parts = []
            with ZipFile(self.path) as zf:
                for name in zf.namelist():
                    upath = normalize_path(name)
                    parts.append(upath)
            return parts
        elif self.__packaging == FOLDER:
            return self._get_folder_parts()
        elif self.__packaging == XML:
            # For flat XML, parts are stored in memory
            return list(self.__parts.keys())
        else:
            raise ValueError("Unable to provide parts of the document")

    @property
    def parts(self) -> list[str]:
        """Get the list of the parts in the Container.

        Returns:
            The list of path of the parts in the Container.
        """
        return self.get_parts()

    def get_part(self, path: str) -> str | bytes | None:
        """Get the actual content of a part of the ODF Container.

        Args:
            path: path of the required part.

        Returns:
            The actual content of the part.
        """
        path = str(path)
        if path in self.__parts:
            part = self.__parts[path]
            if part is None:
                raise ValueError(f'Part "{path}" is deleted')
            if self.__packaging == FOLDER:
                cache_ts = self.__parts_ts.get(path, -1)
                current_ts = self._get_folder_part_timestamp(path)
                if current_ts != cache_ts:
                    part, timestamp = self._get_folder_part(path)
                    self.__parts[path] = part
                    self.__parts_ts[path] = timestamp
            return part
        if self.__packaging == ZIP:
            return self._get_zip_part(path)
        if self.__packaging == FOLDER:
            part, timestamp = self._get_folder_part(path)
            self.__parts[path] = part
            self.__parts_ts[path] = timestamp
            return part
        return None

    @property
    def default_manifest_rdf(self) -> str:
        return (
            '<?xml version="1.0" encoding="utf-8"?>\n'
            '<rdf:RDF xmlns:rdf="http://www.w3.org/'
            '1999/02/22-rdf-syntax-ns#">\n'
            '  <rdf:Description rdf:about="styles.xml">\n'
            '    <rdf:type rdf:resource="http://docs.oasis-open.org/'
            f'ns/office/{OFFICE_VERSION}/meta/odf#StylesFile"/>\n'
            "  </rdf:Description>\n"
            '  <rdf:Description rdf:about="">\n'
            '    <ns0:hasPart xmlns:ns0="http://docs.oasis-open.org/ns/'
            f'office/{OFFICE_VERSION}/meta/pkg#" rdf:resource="styles.xml"/>\n'
            "  </rdf:Description>\n"
            '  <rdf:Description rdf:about="content.xml">\n'
            '    <rdf:type rdf:resource="http://docs.oasis-open.org/'
            f'ns/office/{OFFICE_VERSION}/meta/odf#ContentFile"/>\n'
            "  </rdf:Description>\n"
            '  <rdf:Description rdf:about="">\n'
            '    <ns0:hasPart xmlns:ns0="http://docs.oasis-open.org/ns/office'
            f'/{OFFICE_VERSION}/meta/pkg#" rdf:resource="content.xml"/>\n'
            "  </rdf:Description>\n"
            '  <rdf:Description rdf:about="">\n'
            '    <rdf:type rdf:resource="http://docs.oasis-open.org/ns/office'
            f'/{OFFICE_VERSION}/meta/pkg#Document"/>\n'
            "  </rdf:Description>\n"
            "</rdf:RDF>\n"
        )

    @property
    def mimetype(self) -> str:
        """Return str value of mimetype of the document."""
        with contextlib.suppress(Exception):
            b_mimetype = self.get_part("mimetype")
            if isinstance(b_mimetype, bytes):
                return bytes_to_str(b_mimetype)
        return ""

    @mimetype.setter
    def mimetype(self, mimetype: str | bytes) -> None:
        """Set mimetype value of the document."""
        if isinstance(mimetype, str):
            self.__parts["mimetype"] = str_to_bytes(mimetype)
        elif isinstance(mimetype, bytes):
            self.__parts["mimetype"] = mimetype
        else:
            raise TypeError(f'Wrong mimetype "{mimetype!r}"')

    def set_part(self, path: str, data: bytes) -> None:
        """Replace or add a new part.

        Args:
            path: The relative path in the Container.
            data: Content of the part.
        """
        self.__parts[path] = data

    def del_part(self, path: str) -> None:
        """Mark a part for deletion.

        Args:
            path: The relative path in the Container.
        """
        self.__parts[path] = None

    @property
    def clone(self) -> Container:
        """Make a copy of this container with no path."""
        if self.path and self.__packaging == ZIP:
            self._get_all_zip_part()
        clone = deepcopy(self)
        clone.path = None
        return clone

    def _backup_or_unlink(self, backup: bool, target: str | Path) -> None:
        if backup:
            self._do_backup(target)
        else:
            self._do_unlink(target)

    @staticmethod
    def _do_backup(target: str | Path) -> None:
        path = Path(target)
        if not path.exists():
            return
        back_file = Path(path.stem + ".backup" + path.suffix)
        if back_file.is_dir():
            try:
                shutil.rmtree(back_file)
            except OSError as e:
                printwarn(str(e))
        try:
            shutil.move(target, back_file)
        except OSError as e:
            printwarn(str(e))

    @staticmethod
    def _do_unlink(target: str | Path) -> None:
        path = Path(target)
        if path.exists():
            try:
                shutil.rmtree(path)
            except OSError as e:
                printwarn(str(e))

    def _clean_save_packaging(self, packaging: str | None) -> str:
        if not packaging:
            packaging = self.__packaging if self.__packaging else ZIP
        packaging = packaging.strip().lower()
        if packaging not in PACKAGING:
            msg = f'Packaging of type "{packaging}" is not supported'
            raise ValueError(msg)
        return packaging

    def _clean_save_target(
        self,
        target: str | Path | io.BytesIO | None,
    ) -> str | io.BytesIO:
        if target is None:
            target = self.path
        if isinstance(target, Path):
            target = str(target)
        if isinstance(target, str):
            while target.endswith(os.sep):
                target = target[:-1]
            while target.endswith(".folder"):
                target = target.split(".folder", 1)[0]
        return target  # type: ignore

    def _save_as_zip(
        self,
        target: str | Path | io.BytesIO,
        backup: bool,
    ) -> None:
        if isinstance(target, (str, Path)) and backup:
            self._do_backup(target)
        self._save_zip(target)

    def _save_as_folder(self, target: str | Path, backup: bool) -> None:
        if not isinstance(target, (str, Path)):
            raise TypeError(
                f"Saving in folder format requires a folder name, not '{target!r}'"
            )
        if not str(target).endswith(".folder"):
            target = str(target) + ".folder"
        self._backup_or_unlink(backup, target)
        self._save_folder(target)

    def _save_as_xml(
        self,
        target: str | Path | io.BytesIO,
        backup: bool,
        pretty: bool = True,
    ) -> None:
        if not isinstance(target, (str, Path, io.BytesIO)):
            raise TypeError(
                f"Saving in XML format requires a path name, not '{target!r}'"
            )
        if isinstance(target, (str, Path)):
            target_path = Path(target)
            suffix = target_path.suffix.lower()
            if suffix not in ODF_FLAT_EXTENSIONS and suffix != ".xml":
                flat_ext = ODF_MIMETYPE_TO_FLAT_EXTENSION.get(self.mimetype, ".xml")
                target_path = target_path.with_suffix(flat_ext)
            if backup:
                self._do_backup(target_path)
            self._save_xml(target_path, pretty)
        else:
            self._save_xml(target, pretty)

    def save(
        self,
        target: str | Path | io.BytesIO | None,
        packaging: str | None = None,
        backup: bool = False,
        pretty: bool = False,
    ) -> None:
        """Save the container to the given target, a path or a file-like
        object.

        Package the output document in the same format than current document,
        unless "packaging" is different.

        Args:
            target: The path, file-like object, or Path object where the
                document will be saved.
            packaging: The packaging format to use. Can be 'zip', 'xml', or
                'folder'. If None, the current packaging is used.
            backup: If True, a backup of the original file is created.
            pretty: If True, the XML output will be pretty-printed.
        """
        parts = self.__parts
        packaging = self._clean_save_packaging(packaging)
        # Load parts else they will be considered deleted
        for path in self.parts:
            if path not in parts:
                self.get_part(path)
        target = self._clean_save_target(target)
        if packaging == FOLDER:
            if isinstance(target, io.BytesIO):
                msg = "Impossible to save on io.BytesIO with 'folder' packaging"
                raise TypeError(msg)
            self._save_as_folder(target, backup)
        elif packaging == XML:
            self._save_as_xml(target, backup, pretty)
        else:
            # default:
            self._save_as_zip(target, backup)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormFile


@pytest.fixture
def form_file() -> Iterable[FormFile]:
    yield FormFile(
        name="some name",
        control_implementation="some implem",
        value="initial",
        current_value="final",
        disabled=False,
        printable=True,
        readonly=False,
        tab_index=4,
        tab_stop=False,
        xml_id="control1",
    )


def test_form_file_class():
    form = FormFile()
    assert isinstance(form, FormFile)


def test_form_file_minimal_tag():
    form = Element.from_tag("<form:file/>")
    assert isinstance(form, FormFile)


def test_form_file_serialize():
    form = FormFile()
    assert form.serialize() == "<form:file/>"


def test_form_file_repr():
    form = FormFile()
    assert repr(form) == "<FormFile name=None xml_id=None>"


def test_form_file_name(form_file):
    assert form_file.name == "some name"


def test_form_file_control_implementation(form_file):
    assert form_file.control_implementation == "some implem"


def test_form_file_disabled(form_file):
    assert form_file.disabled is False


def test_form_file_printable(form_file):
    assert form_file.printable is True


def test_form_file_tab_index(form_file):
    assert form_file.tab_index == 4


def test_form_file_tab_stop(form_file):
    assert form_file.tab_stop is False


def test_form_file_xml_id(form_file):
    assert form_file.xml_id == "control1"


def test_form_file_form_id(form_file):
    assert form_file.form_id == "control1"


def test_form_file_xforms_bind():
    form = FormFile(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"


def test_form_file_form_id_2():
    form = FormFile(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"


def test_form_file_form_value(form_file):
    assert form_file.value == "initial"


def test_form_file_form_as_dict(form_file):
    expected = {
        "tag": "form:file",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": "final",
        "str": "final",
    }
    assert form_file.as_dict() == expected


def test_form_file_form_as_dict2(form_file):
    form_file.current_value = None
    expected = {
        "tag": "form:file",
        "name": "some name",
        "xml_id": "control1",
        "value": "initial",
        "current_value": None,
        "str": "initial",
    }
    assert form_file.as_dict() == expected


def test_form_file_linked_cell(form_file):
    assert form_file.linked_cell is None


def test_form_file_linked_cell_2():
    form = FormFile(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        linked_cell="a1",
    )
    assert form.linked_cell == "a1"


def test_form_file_readonly(form_file):
    assert form_file.readonly is False


def test_form_file_form_current_value(form_file):
    assert form_file.current_value == "final"


def test_form_file_form_current_value_2(form_file):
    form_file.current_value = None
    assert form_file.current_value is None
